# Copyright: (c) 2022, Matt Martz <matt@sivel.net>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

import os
import pathlib
import sys

here = pathlib.Path(__file__).parent


def _infect():
    os.environ['ANSIBLE_DEVEL_WARNING'] = '0'
    ansible = here / 'ansible' / 'lib'
    ansible_test = here / 'ansible' / 'test' / 'lib'
    sys.path[0:0] = [str(ansible), str(ansible_test)]


_infect()



from __future__ import annotations

import dataclasses
import datetime
import typing as t

from ansible.module_utils.common import messages as _messages
from ansible.module_utils._internal import _datatag


@dataclasses.dataclass(**_datatag._tag_dataclass_kwargs)
class Deprecated(_datatag.AnsibleDatatagBase):
    msg: str
    help_text: t.Optional[str] = None
    removal_date: t.Optional[datetime.date] = None
    removal_version: t.Optional[str] = None
    plugin: t.Optional[_messages.PluginInfo] = None

    @classmethod
    def _from_dict(cls, d: t.Dict[str, t.Any]) -> Deprecated:
        source = d
        removal_date = source.get('removal_date')

        if removal_date is not None:
            source = source.copy()
            source['removal_date'] = datetime.date.fromisoformat(removal_date)

        return cls(**source)

    def _as_dict(self) -> t.Dict[str, t.Any]:
        value = _datatag.AnsibleDatatagBase._as_dict(self)

        if self.removal_date is not None:
            value['removal_date'] = self.removal_date.isoformat()

        return value



# -*- coding: utf-8 -*-
# (c) 2019, Jordan Borean <jborean@redhat.com>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

import re
import typing as t

from unittest.mock import Mock

import pytest
import pytest_mock

from ansible.executor.interpreter_discovery import discover_interpreter
from ansible.module_utils.common.text.converters import to_text
from ansible.errors import AnsibleConnectionFailure


mock_ubuntu_platform_res = to_text(
    r'{"osrelease_content": "NAME=\"Ansible Test\"\nVERSION=\"100\"\nID=ansible-test\nID_LIKE=debian\n'
    r'PRETTY_NAME=\"Ansible Test 100\"\nVERSION_ID=\"100\"\nHOME_URL=\"http://ansible.com/\"\n'
    r'SUPPORT_URL=\"http://github.com/ansible/ansible\"\nBUG_REPORT_URL=\"http://github.com/ansible/ansible/\"\n'
    r'VERSION_CODENAME=beans\nUBUNTU_CODENAME=beans\n", "platform_dist_result": ["Ansible Test", "100", "beans"]}'
)


@pytest.fixture
def mock_display(mocker: pytest_mock.MockerFixture) -> t.Iterator[Mock]:
    yield mocker.patch('ansible.executor.interpreter_discovery.display', Mock())


def test_discovery_interpreter_linux_auto_legacy(mock_display: Mock) -> None:
    res1 = u'PLATFORM\nLinux\nFOUND\n/usr/bin/python99\n/usr/bin/python3\nENDFOUND'

    mock_action = Mock()
    mock_action._low_level_execute_command.side_effect = [{'stdout': res1}, {'stdout': mock_ubuntu_platform_res}]

    actual = discover_interpreter(mock_action, 'python', 'auto_legacy', {'inventory_hostname': u'host-fóöbär'})

    assert actual == u'/usr/bin/python3'
    assert len(mock_action.method_calls) > 1
    assert len(mock_display.warning.mock_calls) >= 1

    expected_warning = ("Distribution 'Ansible Test' '100' on host 'host-fóöbär' should use '/usr/bin/python99', "
                        "but is using '/usr/bin/python3' for backward compatibility with prior Ansible releases.")

    # this is broken out to allow pytest assert rewrite to show the candidate values instead of `<genexpr>` on failure
    candidate_msgs = [call.kwargs['msg'] for call in mock_display.warning.mock_calls if 'msg' in call.kwargs]
    assert expected_warning in candidate_msgs


def test_discovery_interpreter_linux_auto_legacy_silent() -> None:
    res1 = u'PLATFORM\nLinux\nFOUND\n/usr/bin/python3.9\n/usr/bin/python3\nENDFOUND'

    mock_action = Mock()
    mock_action._low_level_execute_command.side_effect = [{'stdout': res1}, {'stdout': mock_ubuntu_platform_res}]

    actual = discover_interpreter(mock_action, 'python', 'auto_legacy_silent', {'inventory_hostname': u'host-fóöbär'})

    assert actual == u'/usr/bin/python3'
    assert len(mock_action.method_calls) == 2


def test_discovery_interpreter_linux_auto() -> None:
    res1 = u'PLATFORM\nLinux\nFOUND\n/usr/bin/python99\n/usr/bin/python3\nENDFOUND'

    mock_action = Mock()
    mock_action._low_level_execute_command.side_effect = [{'stdout': res1}, {'stdout': mock_ubuntu_platform_res}]

    actual = discover_interpreter(mock_action, 'python', 'auto', {'inventory_hostname': u'host-fóöbär'})

    assert actual == u'/usr/bin/python99'
    assert len(mock_action.method_calls) == 2


def test_discovery_interpreter_non_linux(mock_display: Mock) -> None:
    mock_action = Mock()
    mock_action._low_level_execute_command.return_value = \
        {'stdout': u'PLATFORM\nDarwin\nFOUND\n/usr/bin/python3\nENDFOUND'}

    actual = discover_interpreter(mock_action, 'python', 'auto_legacy', {'inventory_hostname': u'host-fóöbär'})

    assert actual == u'/usr/bin/python3'
    assert len(mock_action.method_calls) == 1
    expected_warning = ("Platform 'darwin' on host 'host-fóöbär' is using the discovered Python interpreter at '/usr/bin/python3', "
                        "but future installation of another Python interpreter could cause a different interpreter to be discovered.")

    # this is broken out to allow pytest assert rewrite to show the candidate values instead of `<genexpr>` on failure
    candidate_msgs = [call.kwargs['msg'] for call in mock_display.warning.mock_calls if 'msg' in call.kwargs]
    assert expected_warning in candidate_msgs


def test_no_interpreters_found(mock_display: Mock) -> None:
    mock_action = Mock()
    mock_action._low_level_execute_command.return_value = {'stdout': u'PLATFORM\nWindows\nFOUND\nENDFOUND'}

    actual = discover_interpreter(mock_action, 'python', 'auto_legacy', {'inventory_hostname': u'host-fóöbär'})

    assert actual == u'/usr/bin/python3'
    assert len(mock_action.method_calls) == 1

    expected_warning = "No python interpreters found for host 'host-fóöbär'"

    # lop off the part that changes over time, if present
    # this is broken out to allow pytest assert rewrite to show the candidate values instead of `<genexpr>` on failure
    candidate_msgs = [re.sub(r' \(tried .*\)\.$', '', call.kwargs['msg']) for call in mock_display.warning.mock_calls if 'msg' in call.kwargs]
    assert expected_warning in candidate_msgs


def test_ansible_error_exception() -> None:
    mock_action = Mock()
    mock_action._low_level_execute_command.side_effect = AnsibleConnectionFailure("host key mismatch")

    with pytest.raises(AnsibleConnectionFailure) as context:
        discover_interpreter(mock_action, 'python', 'auto_legacy', {'inventory_hostname': u'host'})

    assert 'host key mismatch' == str(context.value)



# Copyright (c) 2017 Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

import os
import time
import typing as t

from ansible import constants as C
from ansible.errors import AnsibleActionFail
from ansible.executor.module_common import _apply_action_arg_defaults
from ansible.module_utils.parsing.convert_bool import boolean
from ansible.plugins.action import ActionBase
from ansible.utils.vars import merge_hash


class ActionModule(ActionBase):

    _supports_check_mode = True

    def _get_module_args(self, fact_module: str, task_vars: dict[str, t.Any]) -> dict[str, t.Any]:

        mod_args = self._task.args.copy()

        # deal with 'setup specific arguments'
        if fact_module not in C._ACTION_SETUP:

            # TODO: remove in favor of controller side argspec detecting valid arguments
            # network facts modules must support gather_subset
            try:
                name = self._connection.ansible_name.removeprefix('ansible.netcommon.')
            except AttributeError:
                name = self._connection._load_name.split('.')[-1]
            if name not in ('network_cli', 'httpapi', 'netconf'):
                subset = mod_args.pop('gather_subset', None)
                if subset not in ('all', ['all'], None):
                    self._display.warning('Not passing subset(%s) to %s' % (subset, fact_module))

            timeout = mod_args.pop('gather_timeout', None)
            if timeout is not None:
                self._display.warning('Not passing timeout(%s) to %s' % (timeout, fact_module))

            fact_filter = mod_args.pop('filter', None)
            if fact_filter is not None:
                self._display.warning('Not passing filter(%s) to %s' % (fact_filter, fact_module))

        # Strip out keys with ``None`` values, effectively mimicking ``omit`` behavior
        # This ensures we don't pass a ``None`` value as an argument expecting a specific type
        mod_args = dict((k, v) for k, v in mod_args.items() if v is not None)

        # handle module defaults
        resolved_fact_module = self._shared_loader_obj.module_loader.find_plugin_with_context(
            fact_module, collection_list=self._task.collections
        ).resolved_fqcn

        mod_args = _apply_action_arg_defaults(resolved_fact_module, self._task, mod_args, self._templar)

        return mod_args

    def _combine_task_result(self, result: dict[str, t.Any], task_result: dict[str, t.Any]) -> dict[str, t.Any]:
        """ builds the final result to return """
        filtered_res = {
            'ansible_facts': task_result.get('ansible_facts', {}),
            'warnings': task_result.get('warnings', []),
            'deprecations': task_result.get('deprecations', []),
        }

        # on conflict the last plugin processed wins, but try to do deep merge and append to lists.
        return merge_hash(result, filtered_res, list_merge='append_rp')

    def _handle_smart(self, modules: list, task_vars: dict[str, t.Any]):
        """ Updates the module list when 'smart' is used, lookup network os mappings or use setup, warn when things seem inconsistent """

        if 'smart' not in modules:
            return

        modules.pop(modules.index('smart'))  # remove as this will cause 'module not found' errors
        network_os = self._task.args.get('network_os', task_vars.get('ansible_network_os', task_vars.get('ansible_facts', {}).get('network_os')))

        if network_os:

            connection_map = C.config.get_config_value('CONNECTION_FACTS_MODULES', variables=task_vars)
            if network_os in connection_map:
                modules.append(connection_map[network_os])
            elif not modules:
                raise AnsibleActionFail(f"No fact modules available and we could not find a fact module for your network OS ({network_os}), "
                                        "try setting one via the `FACTS_MODULES` configuration.")

            if set(modules).intersection(set(C._ACTION_SETUP)):
                # most don't realize how setup works with networking connection plugins (forced_local)
                self._display.warning("Detected 'setup' module and a network OS is set, the output when running it will reflect 'localhost'"
                                      " and not the target when a networking connection plugin is used.")

        elif not set(modules).difference(set(C._ACTION_SETUP)):
            # no network OS and setup not in list, add setup by default since 'smart'
            modules.append('ansible.legacy.setup')

    def run(self, tmp: t.Optional[str] = None, task_vars: t.Optional[dict[str, t.Any]] = None) -> dict[str, t.Any]:

        result = super(ActionModule, self).run(tmp, task_vars)
        result['ansible_facts'] = {}

        # copy the value with list() so we don't mutate the config
        modules = list(C.config.get_config_value('FACTS_MODULES', variables=task_vars))
        self._handle_smart(modules, task_vars)

        parallel = task_vars.pop('ansible_facts_parallel', self._task.args.pop('parallel', None))

        failed = {}
        skipped = {}

        if parallel is None:
            if len(modules) > 1:
                parallel = True
            else:
                parallel = False
        else:
            parallel = boolean(parallel)

        timeout = self._task.args.get('gather_timeout', None)
        async_val = self._task.async_val

        if not parallel:
            # serially execute each module
            for fact_module in modules:
                # just one module, no need for fancy async
                mod_args = self._get_module_args(fact_module, task_vars)
                # TODO: use gather_timeout to cut module execution if module itself does not support gather_timeout
                res = self._execute_module(module_name=fact_module, module_args=mod_args, task_vars=task_vars, wrap_async=False)
                if res.get('failed', False):
                    # DTFIX-RELEASE: this trashes the individual failure details and does not work with the new error handling; need to do something to
                    # invoke per-item error handling- perhaps returning this as a synthetic loop result?
                    failed[fact_module] = res
                elif res.get('skipped', False):
                    skipped[fact_module] = res
                else:
                    result = self._combine_task_result(result, res)

            self._remove_tmp_path(self._connection._shell.tmpdir)
        else:
            # do it async, aka parallel
            jobs = {}

            for fact_module in modules:
                mod_args = self._get_module_args(fact_module, task_vars)

                #  if module does not handle timeout, use timeout to handle module, hijack async_val as this is what async_wrapper uses
                # TODO: make this action complain about async/async settings, use parallel option instead .. or remove parallel in favor of async settings?
                if timeout and 'gather_timeout' not in mod_args:
                    self._task.async_val = int(timeout)
                elif async_val != 0:
                    self._task.async_val = async_val
                else:
                    self._task.async_val = 0

                self._display.vvvv("Running %s" % fact_module)
                jobs[fact_module] = (self._execute_module(module_name=fact_module, module_args=mod_args, task_vars=task_vars, wrap_async=True))

            while jobs:
                for module in jobs:
                    poll_args = {'jid': jobs[module]['ansible_job_id'], '_async_dir': os.path.dirname(jobs[module]['results_file'])}
                    res = self._execute_module(module_name='ansible.legacy.async_status', module_args=poll_args, task_vars=task_vars, wrap_async=False)
                    if res.get('finished', 0) == 1:
                        if res.get('failed', False):
                            # DTFIX-RELEASE: this trashes the individual failure details and does not work with the new error handling; need to do something to
                            # invoke per-item error handling- perhaps returning this as a synthetic loop result?
                            failed[module] = res
                        elif res.get('skipped', False):
                            skipped[module] = res
                        else:
                            result = self._combine_task_result(result, res)
                        del jobs[module]
                        break
                    else:
                        time.sleep(0.1)
                else:
                    time.sleep(0.5)

        # restore value for post processing
        if self._task.async_val != async_val:
            self._task.async_val = async_val

        if skipped:
            result['msg'] = "The following modules were skipped: %s\n" % (', '.join(skipped.keys()))
            result['skipped_modules'] = skipped
            if len(skipped) == len(modules):
                result['skipped'] = True

        if failed:
            result['failed'] = True
            result['msg'] = "The following modules failed to execute: %s\n" % (', '.join(failed.keys()))
            result['failed_modules'] = failed

        # tell executor facts were gathered
        result['ansible_facts']['_ansible_facts_gathered'] = True

        # hack to keep --verbose from showing all the setup module result
        result['_ansible_verbose_override'] = True

        return result



"""Patch broken ClassVar support in dataclasses when ClassVar is accessed via a module other than `typing`."""

# deprecated: description='verify ClassVar support in dataclasses has been fixed in Python before removing this patching code', python_version='3.12'

from __future__ import annotations

import dataclasses
import sys
import typing as t

# trigger the bug by exposing typing.ClassVar via a module reference that is not `typing`
_ts = sys.modules[__name__]
ClassVar = t.ClassVar


def patch_dataclasses_is_type() -> None:
    if not _is_patch_needed():
        return  # pragma: nocover

    try:
        real_is_type = dataclasses._is_type  # type: ignore[attr-defined]
    except AttributeError:  # pragma: nocover
        raise RuntimeError("unable to patch broken dataclasses ClassVar support") from None

    # patch dataclasses._is_type - impl from https://github.com/python/cpython/blob/4c6d4f5cb33e48519922d635894eef356faddba2/Lib/dataclasses.py#L709-L765
    def _is_type(annotation, cls, a_module, a_type, is_type_predicate):
        match = dataclasses._MODULE_IDENTIFIER_RE.match(annotation)  # type: ignore[attr-defined]
        if match:
            ns = None
            module_name = match.group(1)
            if not module_name:
                # No module name, assume the class's module did
                # "from dataclasses import InitVar".
                ns = sys.modules.get(cls.__module__).__dict__
            else:
                # Look up module_name in the class's module.
                module = sys.modules.get(cls.__module__)
                if module and module.__dict__.get(module_name):  # this is the patched line; removed `is a_module`
                    ns = sys.modules.get(a_type.__module__).__dict__
            if ns and is_type_predicate(ns.get(match.group(2)), a_module):
                return True
        return False

    _is_type._orig_impl = real_is_type  # type: ignore[attr-defined]  # stash this away to allow unit tests to undo the patch

    dataclasses._is_type = _is_type  # type: ignore[attr-defined]

    try:
        if _is_patch_needed():
            raise RuntimeError("patching had no effect")  # pragma: nocover
    except Exception as ex:  # pragma: nocover
        dataclasses._is_type = real_is_type  # type: ignore[attr-defined]
        raise RuntimeError("dataclasses ClassVar support is still broken after patching") from ex


def _is_patch_needed() -> bool:
    @dataclasses.dataclass
    class CheckClassVar:
        # this is the broken case requiring patching: ClassVar dot-referenced from a module that is not `typing` is treated as an instance field
        # DTFIX-RELEASE: add link to CPython bug report to-be-filed (or update associated deprecation comments if we don't)
        a_classvar: _ts.ClassVar[int]  # type: ignore[name-defined]
        a_field: int

    return len(dataclasses.fields(CheckClassVar)) != 1



from __future__ import annotations

from jinja2 import __version__ as _jinja2_version

# DTFIX-FUTURE: sanity test to ensure this doesn't drift from requirements
_MINIMUM_JINJA_VERSION = (3, 1)

if tuple(map(int, _jinja2_version.split('.', maxsplit=2)[:2])) < _MINIMUM_JINJA_VERSION:
    raise RuntimeError(f'Jinja version {".".join(map(str, _MINIMUM_JINJA_VERSION))} or higher is required (current version {_jinja2_version}).')



# Copyright: Contributors to the Ansible project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

import os

from ansible.module_utils.facts.virtual.base import Virtual, VirtualCollector


class SunOSVirtual(Virtual):
    """
    This is a SunOS-specific subclass of Virtual.  It defines
    - virtualization_type
    - virtualization_role
    - container
    """
    platform = 'SunOS'

    def get_virtual_facts(self):
        virtual_facts = {}
        host_tech = set()
        guest_tech = set()

        # Check if it's a zone
        zonename = self.module.get_bin_path('zonename')
        if zonename:
            rc, out, err = self.module.run_command(zonename)
            if rc == 0:
                if out.rstrip() == "global":
                    host_tech.add('zone')
                else:
                    guest_tech.add('zone')
                    virtual_facts['container'] = 'zone'

        # Check if it's a branded zone (i.e. Solaris 8/9 zone)
        if os.path.isdir('/.SUNWnative'):
            guest_tech.add('zone')
            virtual_facts['container'] = 'zone'

        # If it's a zone check if we can detect if our global zone is itself virtualized.
        # Relies on the "guest tools" (e.g. vmware tools) to be installed
        if 'container' in virtual_facts and virtual_facts['container'] == 'zone':
            modinfo = self.module.get_bin_path('modinfo')
            if modinfo:
                rc, out, err = self.module.run_command(modinfo)
                if rc == 0:
                    for line in out.splitlines():
                        if 'VMware' in line:
                            guest_tech.add('vmware')
                            virtual_facts['virtualization_type'] = 'vmware'
                            virtual_facts['virtualization_role'] = 'guest'
                        if 'VirtualBox' in line:
                            guest_tech.add('virtualbox')
                            virtual_facts['virtualization_type'] = 'virtualbox'
                            virtual_facts['virtualization_role'] = 'guest'

        if os.path.exists('/proc/vz'):
            guest_tech.add('virtuozzo')
            virtual_facts['virtualization_type'] = 'virtuozzo'
            virtual_facts['virtualization_role'] = 'guest'

        # Detect domaining on Sparc hardware
        virtinfo = self.module.get_bin_path('virtinfo')
        if virtinfo:
            # The output of virtinfo is different whether we are on a machine with logical
            # domains ('LDoms') on a T-series or domains ('Domains') on a M-series. Try LDoms first.
            rc, out, err = self.module.run_command([virtinfo, '-p'])
            # The output contains multiple lines with different keys like this:
            #   DOMAINROLE|impl=LDoms|control=false|io=false|service=false|root=false
            # The output may also be not formatted and the returncode is set to 0 regardless of the error condition:
            #   virtinfo can only be run from the global zone
            if rc == 0:
                try:
                    for line in out.splitlines():
                        fields = line.split('|')
                        if fields[0] == 'DOMAINROLE' and fields[1] == 'impl=LDoms':
                            guest_tech.add('ldom')
                            virtual_facts['virtualization_type'] = 'ldom'
                            virtual_facts['virtualization_role'] = 'guest'
                            hostfeatures = []
                            for field in fields[2:]:
                                arg = field.split('=')
                                if arg[1] == 'true':
                                    hostfeatures.append(arg[0])
                            if len(hostfeatures) > 0:
                                virtual_facts['virtualization_role'] = 'host (' + ','.join(hostfeatures) + ')'
                except ValueError:
                    pass

        else:
            smbios = self.module.get_bin_path('smbios')
            if not smbios:
                return
            rc, out, err = self.module.run_command(smbios)
            if rc == 0:
                for line in out.splitlines():
                    if 'VMware' in line:
                        guest_tech.add('vmware')
                        virtual_facts['virtualization_type'] = 'vmware'
                        virtual_facts['virtualization_role'] = 'guest'
                    elif 'Parallels' in line:
                        guest_tech.add('parallels')
                        virtual_facts['virtualization_type'] = 'parallels'
                        virtual_facts['virtualization_role'] = 'guest'
                    elif 'VirtualBox' in line:
                        guest_tech.add('virtualbox')
                        virtual_facts['virtualization_type'] = 'virtualbox'
                        virtual_facts['virtualization_role'] = 'guest'
                    elif 'HVM domU' in line:
                        guest_tech.add('xen')
                        virtual_facts['virtualization_type'] = 'xen'
                        virtual_facts['virtualization_role'] = 'guest'
                    elif 'KVM' in line:
                        guest_tech.add('kvm')
                        virtual_facts['virtualization_type'] = 'kvm'
                        virtual_facts['virtualization_role'] = 'guest'

        virtual_facts['virtualization_tech_guest'] = guest_tech
        virtual_facts['virtualization_tech_host'] = host_tech
        return virtual_facts


class SunOSVirtualCollector(VirtualCollector):
    _fact_class = SunOSVirtual
    _platform = 'SunOS'



from __future__ import annotations

import typing as t

from ._ambient_context import AmbientContextBase
from ..common.messages import PluginInfo


class HasPluginInfo(t.Protocol):
    """Protocol to type-annotate and expose PluginLoader-set values."""

    @property
    def _load_name(self) -> str:
        """The requested name used to load the plugin."""

    @property
    def ansible_name(self) -> str:
        """Fully resolved plugin name."""

    @property
    def plugin_type(self) -> str:
        """Plugin type name."""


class PluginExecContext(AmbientContextBase):
    """Execution context that wraps all plugin invocations to allow infrastructure introspection of the currently-executing plugin instance."""

    def __init__(self, executing_plugin: HasPluginInfo) -> None:
        self._executing_plugin = executing_plugin

    @property
    def executing_plugin(self) -> HasPluginInfo:
        return self._executing_plugin

    @property
    def plugin_info(self) -> PluginInfo:
        return PluginInfo(
            requested_name=self._executing_plugin._load_name,
            resolved_name=self._executing_plugin.ansible_name,
            type=self._executing_plugin.plugin_type,
        )

    @classmethod
    def get_current_plugin_info(cls) -> PluginInfo | None:
        """Utility method to extract a PluginInfo for the currently executing plugin (or None if no plugin is executing)."""
        if ctx := cls.current(optional=True):
            return ctx.plugin_info

        return None



# Copyright (c) 2024 Ansible Project
# Simplified BSD License (see licenses/simplified_bsd.txt or https://opensource.org/licenses/BSD-2-Clause)

"""Support code for exclusive use by the AnsiballZ wrapper."""

from __future__ import annotations

import atexit
import dataclasses
import importlib.util
import json
import os
import runpy
import sys
import typing as t

from . import _errors
from ._plugin_exec_context import PluginExecContext, HasPluginInfo
from .. import basic
from ..common.json import AnsibleJSONEncoder
from ..common.messages import PluginInfo


def run_module(
    *,
    json_params: bytes,
    profile: str,
    plugin_info_dict: dict[str, object],
    module_fqn: str,
    modlib_path: str,
    init_globals: dict[str, t.Any] | None = None,
    coverage_config: str | None = None,
    coverage_output: str | None = None,
) -> None:  # pragma: nocover
    """Used internally by the AnsiballZ wrapper to run an Ansible module."""
    try:
        _enable_coverage(coverage_config, coverage_output)
        _run_module(
            json_params=json_params,
            profile=profile,
            plugin_info_dict=plugin_info_dict,
            module_fqn=module_fqn,
            modlib_path=modlib_path,
            init_globals=init_globals,
        )
    except Exception as ex:  # not BaseException, since modules are expected to raise SystemExit
        _handle_exception(ex)


def _enable_coverage(coverage_config: str | None, coverage_output: str | None) -> None:  # pragma: nocover
    """Bootstrap `coverage` for the current Ansible module invocation."""
    if not coverage_config:
        return

    if coverage_output:
        # Enable code coverage analysis of the module.
        # This feature is for internal testing and may change without notice.
        python_version_string = '.'.join(str(v) for v in sys.version_info[:2])
        os.environ['COVERAGE_FILE'] = f'{coverage_output}=python-{python_version_string}=coverage'

        import coverage

        cov = coverage.Coverage(config_file=coverage_config)

        def atexit_coverage():
            cov.stop()
            cov.save()

        atexit.register(atexit_coverage)

        cov.start()
    else:
        # Verify coverage is available without importing it.
        # This will detect when a module would fail with coverage enabled with minimal overhead.
        if importlib.util.find_spec('coverage') is None:
            raise RuntimeError('Could not find the `coverage` Python module.')


def _run_module(
    *,
    json_params: bytes,
    profile: str,
    plugin_info_dict: dict[str, object],
    module_fqn: str,
    modlib_path: str,
    init_globals: dict[str, t.Any] | None = None,
) -> None:
    """Used internally by `_run_module` to run an Ansible module after coverage has been enabled (if applicable)."""
    basic._ANSIBLE_ARGS = json_params
    basic._ANSIBLE_PROFILE = profile

    init_globals = init_globals or {}
    init_globals.update(_module_fqn=module_fqn, _modlib_path=modlib_path)

    with PluginExecContext(_ModulePluginWrapper(PluginInfo._from_dict(plugin_info_dict))):
        # Run the module. By importing it as '__main__', it executes as a script.
        runpy.run_module(mod_name=module_fqn, init_globals=init_globals, run_name='__main__', alter_sys=True)

        # An Ansible module must print its own results and exit. If execution reaches this point, that did not happen.
        raise RuntimeError('New-style module did not handle its own exit.')


def _handle_exception(exception: BaseException) -> t.NoReturn:
    """Handle the given exception."""
    result = dict(
        failed=True,
        exception=_errors.create_error_summary(exception),
    )

    print(json.dumps(result, cls=AnsibleJSONEncoder, preserve_datatags=True))  # pylint: disable=ansible-bad-function

    sys.exit(1)  # pylint: disable=ansible-bad-function


@dataclasses.dataclass(frozen=True)
class _ModulePluginWrapper(HasPluginInfo):
    """Modules aren't plugin instances; this adapter implements the `HasPluginInfo` protocol to allow `PluginExecContext` infra to work with modules."""

    plugin: PluginInfo

    @property
    def _load_name(self) -> str:
        return self.plugin.requested_name

    @property
    def ansible_name(self) -> str:
        return self.plugin.resolved_name

    @property
    def plugin_type(self) -> str:
        return self.plugin.type



#!/usr/bin/env python
# PYTHON_ARGCOMPLETE_OK
"""Generate frozen sanity test requirements from source requirements files."""

from __future__ import annotations

import argparse
import dataclasses
import pathlib
import re
import subprocess
import tempfile
import typing as t
import venv

import packaging.version
import packaging.specifiers
import packaging.requirements

try:
    import argcomplete
except ImportError:
    argcomplete = None


FILE = pathlib.Path(__file__).resolve()
ROOT = FILE.parent.parent
SELF = FILE.relative_to(ROOT)


@dataclasses.dataclass(frozen=True)
class SanityTest:
    name: str
    requirements_path: pathlib.Path
    source_path: pathlib.Path

    def freeze_requirements(self) -> None:
        source_requirements = [packaging.requirements.Requirement(re.sub(' #.*$', '', line)) for line in self.source_path.read_text().splitlines()]

        install_packages = {requirement.name for requirement in source_requirements}
        exclude_packages = {'distribute', 'pip', 'setuptools', 'wheel'} - install_packages

        with tempfile.TemporaryDirectory() as venv_dir:
            venv.create(venv_dir, with_pip=True)

            python = pathlib.Path(venv_dir, 'bin', 'python')
            pip = [python, '-m', 'pip', '--disable-pip-version-check']
            env = dict()

            pip_freeze = subprocess.run(pip + ['freeze'], env=env, check=True, capture_output=True, text=True)

            if pip_freeze.stdout:
                raise Exception(f'Initial virtual environment is not empty:\n{pip_freeze.stdout}')

            subprocess.run(pip + ['install', '-r', self.source_path], env=env, check=True)

            freeze_options = ['--all']

            for exclude_package in exclude_packages:
                freeze_options.extend(('--exclude', exclude_package))

            pip_freeze = subprocess.run(pip + ['freeze'] + freeze_options, env=env, check=True, capture_output=True, text=True)

        self.write_requirements(pip_freeze.stdout)

    def update_pre_build(self) -> None:
        """Update requirements in place with current pre-build instructions."""
        requirements = pathlib.Path(self.requirements_path).read_text()
        lines = requirements.splitlines(keepends=True)
        lines = [line for line in lines if not line.startswith('#')]
        requirements = ''.join(lines)

        self.write_requirements(requirements)

    def write_requirements(self, requirements: str) -> None:
        """Write the given test requirements to the requirements file for this test."""
        pre_build = pre_build_instructions(requirements)

        requirements = f'# edit "{self.source_path.name}" and generate with: {SELF} --test {self.name}\n{pre_build}{requirements}'

        with open(self.requirements_path, 'w') as requirement_file:
            requirement_file.write(requirements)

    @staticmethod
    def create(path: pathlib.Path) -> SanityTest:
        return SanityTest(
            name=path.stem.replace('sanity.', '').replace('.requirements', ''),
            requirements_path=path,
            source_path=path.with_suffix('.in'),
        )


def pre_build_instructions(requirements: str) -> str:
    """Parse the given requirements and return any applicable pre-build instructions."""
    parsed_requirements = requirements.splitlines()

    package_versions = {
        match.group('package').lower(): match.group('version') for match
        in (re.search('^(?P<package>.*)==(?P<version>.*)$', requirement) for requirement in parsed_requirements)
        if match
    }

    instructions: list[str] = []

    build_constraints = (
        ('pyyaml', '>= 5.4, <= 6.0', ('Cython < 3.0',)),
    )

    for package, specifier, constraints in build_constraints:
        version_string = package_versions.get(package)

        if version_string:
            version = packaging.version.Version(version_string)
            specifier_set = packaging.specifiers.SpecifierSet(specifier)

            if specifier_set.contains(version):
                instructions.append(f'# pre-build requirement: {package} == {version}\n')

                for constraint in constraints:
                    instructions.append(f'# pre-build constraint: {constraint}\n')

    return ''.join(instructions)


def main() -> None:
    tests = find_tests()

    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--test',
        metavar='TEST',
        dest='test_names',
        action='append',
        choices=[test.name for test in tests],
        help='test requirements to update'
    )

    parser.add_argument(
        '--pre-build-only',
        action='store_true',
        help='apply pre-build instructions to existing requirements',
    )

    if argcomplete:
        argcomplete.autocomplete(parser)

    args = parser.parse_args()
    test_names: set[str] = set(args.test_names or [])

    tests = [test for test in tests if test.name in test_names] if test_names else tests

    for test in tests:
        print(f'===[ {test.name} ]===', flush=True)

        if args.pre_build_only:
            test.update_pre_build()
        else:
            test.freeze_requirements()


def find_tests() -> t.List[SanityTest]:
    globs = (
        'test/lib/ansible_test/_data/requirements/sanity.*.txt',
        'test/sanity/code-smell/*.requirements.txt',
    )

    tests: t.List[SanityTest] = []

    for glob in globs:
        tests.extend(get_tests(pathlib.Path(glob)))

    return sorted(tests, key=lambda test: test.name)


def get_tests(glob: pathlib.Path) -> t.List[SanityTest]:
    path = pathlib.Path(ROOT, glob.parent)
    pattern = glob.name

    return [SanityTest.create(item) for item in path.glob(pattern)]


if __name__ == '__main__':
    main()



from __future__ import annotations

import copy
import typing as t

from ansible._internal._templating._lazy_containers import _AnsibleLazyTemplateMixin
from ansible._internal._templating._engine import TemplateOptions


# DTFIX-MERGE: Ensure Galaxy CLI doesn't show warnings for this collection (eg, does not have a MANIFEST.json file, nor has it galaxy.yml)
def unmask(value: object, type_names: str | list[str]) -> object:
    """
    Internal filter to suppress automatic type transformation in Jinja (e.g., WarningMessageDetail, DeprecationMessageDetail, ErrorDetail).
    Lazy collection caching is in play- the first attempt to access a value in a given lazy container must be with unmasking in place, or the transformed value
    will already be cached.
    """
    if isinstance(type_names, str):
        type_names = [type_names]

    # DTFIX-MERGE: validation

    result: object

    if isinstance(value, _AnsibleLazyTemplateMixin):
        result = copy.copy(value)
        result._template_options = TemplateOptions(unmask_type_names=frozenset(type_names))
    else:
        result = value

    return result


class FilterModule(object):
    @staticmethod
    def filters() -> dict[str, t.Callable]:
        return dict(unmask=unmask)



"""Internal utilities for serialization and deserialization."""

from __future__ import annotations

import collections.abc as c
import typing as t

from ansible.errors import AnsibleVariableTypeError

from ansible.module_utils._internal._datatag import (
    _ANSIBLE_ALLOWED_MAPPING_VAR_TYPES,
    _ANSIBLE_ALLOWED_NON_SCALAR_COLLECTION_VAR_TYPES,
    _ANSIBLE_ALLOWED_VAR_TYPES,
    _AnsibleTaggedStr,
    AnsibleTagHelper,
)
from ansible._internal._templating._lazy_containers import _AnsibleLazyTemplateDict, _AnsibleLazyTemplateList
from ansible.parsing.vault import EncryptedString
from ansible._internal._datatag._tags import Origin, TrustedAsTemplate

_T = t.TypeVar('_T')

_allowed_collection_types: frozenset[type] = _ANSIBLE_ALLOWED_NON_SCALAR_COLLECTION_VAR_TYPES | {_AnsibleLazyTemplateDict, _AnsibleLazyTemplateList}
_allowed_mapping_types: frozenset[type] = _ANSIBLE_ALLOWED_MAPPING_VAR_TYPES | {_AnsibleLazyTemplateDict}
_allowed_var_types: frozenset[type] = _ANSIBLE_ALLOWED_VAR_TYPES | {_AnsibleLazyTemplateDict, _AnsibleLazyTemplateList}

_sentinel = object()


class HasCurrent(t.Protocol):
    """Utility protocol for mixin type safety."""

    _current: t.Any


class StateTrackingMixIn(HasCurrent):
    """Mixin for use with `AnsibleVariableVisitor` to track current visitation context."""

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._stack: list[t.Any] = []

    def __enter__(self) -> None:
        self._stack.append(self._current)

    def __exit__(self, *_args, **_kwargs) -> None:
        self._stack.pop()

    def _get_stack(self) -> list[t.Any]:
        if not self._stack:
            return []

        return self._stack[1:] + [self._current]


class AnsibleVariableVisitor:
    """Utility visitor base class to recursively apply various behaviors and checks to variable object graphs."""

    def __init__(
        self,
        *,
        trusted_as_template: bool = False,
        origin: Origin | None = None,
        convert_mapping_to_dict: bool = False,
        convert_sequence_to_list: bool = False,
        convert_custom_scalars: bool = False,
        allow_encrypted_string: bool = False,
    ):
        super().__init__()  # supports StateTrackingMixIn

        self.trusted_as_template = trusted_as_template
        self.origin = origin
        self.convert_mapping_to_dict = convert_mapping_to_dict
        self.convert_sequence_to_list = convert_sequence_to_list
        self.convert_custom_scalars = convert_custom_scalars
        self.allow_encrypted_string = allow_encrypted_string

        self._current: t.Any = None  # supports StateTrackingMixIn

    def __enter__(self) -> t.Any:
        """No-op context manager dispatcher (delegates to mixin behavior if present)."""
        if func := getattr(super(), '__enter__', None):
            func()

    def __exit__(self, *args, **kwargs) -> t.Any:
        """No-op context manager dispatcher (delegates to mixin behavior if present)."""
        if func := getattr(super(), '__exit__', None):
            func(*args, **kwargs)

    def visit(self, value: _T) -> _T:
        """
        Enforces Ansible's variable type system restrictions before a var is accepted in inventory. Also, conditionally implements template trust
        compatibility, depending on the plugin's declared understanding (or lack thereof). This always recursively copies inputs to fully isolate
        inventory data from what the plugin provided, and prevent any later mutation.
        """
        return self._visit(None, value)

    def _early_visit(self, value, value_type) -> t.Any:
        """Overridable hook point to allow custom string handling in derived visitors."""
        if value_type in (str, _AnsibleTaggedStr):
            # apply compatibility behavior
            if self.trusted_as_template:
                result = TrustedAsTemplate().tag(value)
            else:
                result = value
        else:
            result = _sentinel

        return result

    def _visit(self, key: t.Any, value: _T) -> _T:
        """Internal implementation to recursively visit a data structure's contents."""
        self._current = key  # supports StateTrackingMixIn

        value_type = type(value)

        result: _T

        # DTFIX-MERGE: the visitor is largely ignoring dict/mapping keys, except for debugging and schema-aware checking, it should be doing type checks on keys

        if (result := self._early_visit(value, value_type)) is not _sentinel:
            pass
        elif value_type in _allowed_mapping_types:  # check mappings first, because they're also collections
            with self:  # supports StateTrackingMixIn
                result = AnsibleTagHelper.tag_copy(value, ((k, self._visit(k, v)) for k, v in value.items()), value_type=value_type)
        elif value_type in _allowed_collection_types:
            with self:  # supports StateTrackingMixIn
                result = AnsibleTagHelper.tag_copy(value, (self._visit(k, v) for k, v in enumerate(t.cast(t.Iterable, value))), value_type=value_type)
        elif self.allow_encrypted_string and isinstance(value, EncryptedString):
            return value  # type: ignore[return-value]  # DTFIX-MERGE: this should probably only be allowed for values in dict, not keys (set, dict)
        elif self.convert_mapping_to_dict and isinstance(value, c.Mapping):
            result = {k: self._visit(k, v) for k, v in value.items()}  # type: ignore[assignment]
        # DTFIX-MERGE: share these somehow before they drift (and fix to include set, frozenset)
        elif self.convert_sequence_to_list and isinstance(value, c.Sequence) and not isinstance(value, (str, bytes)):
            result = [self._visit(k, v) for k, v in enumerate(value)]  # type: ignore[assignment]
        # DTFIX-MERGE: share these somehow before they drift and ick
        elif self.convert_custom_scalars and isinstance(value, str):
            result = str(value)  # type: ignore[assignment]
        elif self.convert_custom_scalars and isinstance(value, float):
            result = float(value)  # type: ignore[assignment]
        elif self.convert_custom_scalars and isinstance(value, int) and not isinstance(value, bool):
            result = int(value)  # type: ignore[assignment]
        else:
            if value_type not in _allowed_var_types:
                raise AnsibleVariableTypeError.from_value(obj=value)

            # supported scalar type that requires no special handling, just return as-is
            result = value

        if self.origin and not Origin.is_tagged_on(result):
            # apply shared instance default origin tag
            result = self.origin.tag(result)

        return result



from __future__ import annotations

import datetime as _datetime

from ...module_utils._internal import _datatag
from ...module_utils.common import json as _json
from ...parsing import vault as _vault
from .._datatag import _tags


class _Profile(_json._JSONSerializationProfile):
    """Profile for external cache persistence of inventory/fact data that preserves most tags."""

    serialize_map = {}
    schema_id = 1

    @classmethod
    def post_init(cls, **kwargs):
        cls.allowed_ansible_serializable_types = (
            _json._common_module_types
            | _json._common_module_response_types
            | {
                _datatag._AnsibleTaggedDate,
                _datatag._AnsibleTaggedTime,
                _datatag._AnsibleTaggedDateTime,
                _datatag._AnsibleTaggedStr,
                _datatag._AnsibleTaggedInt,
                _datatag._AnsibleTaggedFloat,
                _datatag._AnsibleTaggedList,
                _datatag._AnsibleTaggedSet,
                _datatag._AnsibleTaggedTuple,
                _datatag._AnsibleTaggedDict,
                _tags.EncryptedSource,
                _tags.Origin,
                _tags.TrustedAsTemplate,
                _vault.EncryptedString,
                _vault.VaultedValue,
            }
        )

        cls.serialize_map = {}
        cls.serialize_map.update(
            {
                set: cls.serialize_as_list,
                tuple: cls.serialize_as_list,
                _datetime.date: _datatag.AnsibleSerializableDate,
                _datetime.time: _datatag.AnsibleSerializableTime,
                _datetime.datetime: _datatag.AnsibleSerializableDateTime,
            }
        )


class Encoder(_json.AnsibleProfileJSONEncoder):
    _profile = _Profile


class Decoder(_json.AnsibleProfileJSONDecoder):
    _profile = _Profile



from __future__ import annotations

import dataclasses
import itertools
import pathlib
import sys
import textwrap
import typing as t

from ansible.module_utils.common.messages import Detail, ErrorSummary
from ansible._internal._datatag._tags import Origin
from ansible.module_utils._internal import _ambient_context, _traceback
from ansible import errors

if t.TYPE_CHECKING:
    from ansible.utils.display import Display


class RedactAnnotatedSourceContext(_ambient_context.AmbientContextBase):
    """
    When active, this context will redact annotated source lines, showing only the origin.
    """


def _dedupe_and_concat_message_chain(message_parts: list[str]) -> str:
    message_parts = list(reversed(message_parts))

    message = message_parts.pop(0)

    for message_part in message_parts:
        # avoid duplicate messages where the cause was already concatenated to the exception message
        if message_part.endswith(message):
            message = message_part
        else:
            message = concat_message(message_part, message)

    return message


def _collapse_error_details(error_details: t.Sequence[Detail]) -> list[Detail]:
    """
    Return a potentially modified error chain, with redundant errors collapsed into previous error(s) in the chain.
    This reduces the verbosity of messages by eliminating repetition when multiple errors in the chain share the same contextual information.
    """
    previous_error = error_details[0]
    previous_warnings: list[str] = []
    collapsed_error_details: list[tuple[Detail, list[str]]] = [(previous_error, previous_warnings)]

    for error in error_details[1:]:
        details_present = error.formatted_source_context or error.help_text
        details_changed = error.formatted_source_context != previous_error.formatted_source_context or error.help_text != previous_error.help_text

        if details_present and details_changed:
            previous_error = error
            previous_warnings = []
            collapsed_error_details.append((previous_error, previous_warnings))
        else:
            previous_warnings.append(error.msg)

    final_error_details: list[Detail] = []

    for error, messages in collapsed_error_details:
        final_error_details.append(dataclasses.replace(error, msg=_dedupe_and_concat_message_chain([error.msg] + messages)))

    return final_error_details


def _get_cause(exception: BaseException) -> BaseException | None:
    # deprecated: description='remove support for orig_exc (should have been deprecated in 2.22)' core_version='2.26'

    if not isinstance(exception, errors.AnsibleError):
        return exception.__cause__

    if exception.__cause__:
        if exception.orig_exc and exception.orig_exc is not exception.__cause__:
            _get_display().warning(
                msg=f"The `orig_exc` argument to `{type(exception).__name__}` was given, but differed from the cause given by `raise ... from`.",
            )

        return exception.__cause__

    if exception.orig_exc:
        # encourage the use of `raise ... from` before deprecating `orig_exc`
        _get_display().warning(msg=f"The `orig_exc` argument to `{type(exception).__name__}` was given without using `raise ... from orig_exc`.")

        return exception.orig_exc

    return None


class _TemporaryDisplay:
    # DTFIX-FUTURE: generalize this and hide it in the display module so all users of Display can benefit

    @staticmethod
    def warning(*args, **kwargs):
        print(f'FALLBACK WARNING: {args} {kwargs}', file=sys.stderr)

    @staticmethod
    def deprecated(*args, **kwargs):
        print(f'FALLBACK DEPRECATION: {args} {kwargs}', file=sys.stderr)


def _get_display() -> Display | _TemporaryDisplay:
    try:
        from ansible.utils.display import Display
    except ImportError:
        return _TemporaryDisplay()

    return Display()


def _create_error_summary(exception: BaseException, event: _traceback.TracebackEvent | None = None) -> ErrorSummary:
    from . import _captured  # avoid circular import due to AnsibleError import

    current_exception: BaseException | None = exception
    error_details: list[Detail] = []

    if event:
        formatted_traceback = _traceback.maybe_extract_traceback(exception, event)
    else:
        formatted_traceback = None

    while current_exception:
        if isinstance(current_exception, errors.AnsibleError):
            include_cause_message = current_exception._include_cause_message
            edc = Detail(
                msg=current_exception._original_message.strip(),
                formatted_source_context=current_exception._formatted_source_context,
                help_text=current_exception._help_text,
            )
        else:
            include_cause_message = True
            edc = Detail(
                msg=str(current_exception).strip(),
            )

        error_details.append(edc)

        if isinstance(current_exception, _captured.AnsibleCapturedError):
            detail = current_exception.error_summary
            error_details.extend(detail.details)

            if formatted_traceback and detail.formatted_traceback:
                formatted_traceback = (
                    f'{detail.formatted_traceback}\n'
                    f'The {current_exception.context} exception above was the direct cause of the following controller exception:\n\n'
                    f'{formatted_traceback}'
                )

        if not include_cause_message:
            break

        current_exception = _get_cause(current_exception)

    return ErrorSummary(details=tuple(error_details), formatted_traceback=formatted_traceback)


def concat_message(left: str, right: str) -> str:
    """Normalize `left` by removing trailing punctuation and spaces before appending new punctuation and `right`."""
    return f'{left.rstrip(". ")}: {right}'


def get_chained_message(exception: BaseException) -> str:
    """
    Return the full chain of exception messages by concatenating the cause(s) until all are exhausted.
    """
    error_summary = _create_error_summary(exception)
    message_parts = [edc.msg for edc in error_summary.details]

    return _dedupe_and_concat_message_chain(message_parts)


@dataclasses.dataclass(kw_only=True, frozen=True)
class SourceContext:
    origin: Origin
    annotated_source_lines: list[str]
    target_line: str | None

    def __str__(self) -> str:
        msg_lines = [f'Origin: {self.origin}']

        if self.annotated_source_lines:
            msg_lines.append('')
            msg_lines.extend(self.annotated_source_lines)

        return '\n'.join(msg_lines)

    @classmethod
    def from_value(cls, value: t.Any) -> SourceContext | None:
        """Attempt to retrieve source and render a contextual indicator from the value's origin (if any)."""
        if value is None:
            return None

        if isinstance(value, Origin):
            origin = value
            value = None
        else:
            origin = Origin.get_tag(value)

        if RedactAnnotatedSourceContext.current(optional=True):
            return cls.error('content redacted')

        if origin and origin.path:
            return cls.from_origin(origin)

        # DTFIX-RELEASE: redaction context may not be sufficient to avoid secret disclosure without SensitiveData and other enhancements
        if value is None:
            truncated_value = None
            annotated_source_lines = []
        else:
            # DTFIX-FUTURE: cleanup/share width
            try:
                value = str(value)
            except Exception as ex:
                value = f'<< context unavailable: {ex} >>'

            truncated_value = textwrap.shorten(value, width=120)
            annotated_source_lines = [truncated_value]

        return SourceContext(
            origin=origin or Origin.UNKNOWN,
            annotated_source_lines=annotated_source_lines,
            target_line=truncated_value,
        )

    @staticmethod
    def error(message: str | None, origin: Origin | None = None) -> SourceContext:
        return SourceContext(
            origin=origin,
            annotated_source_lines=[f'(source not shown: {message})'] if message else [],
            target_line=None,
        )

    @classmethod
    def from_origin(cls, origin: Origin) -> SourceContext:
        """Attempt to retrieve source and render a contextual indicator of an error location."""
        from ansible.parsing.vault import is_encrypted  # avoid circular import

        # DTFIX-FUTURE: support referencing the column after the end of the target line, so we can indicate where a missing character (quote) needs to be added
        #               this is also useful for cases like end-of-stream reported by the YAML parser

        # DTFIX-FUTURE: Implement line wrapping and match annotated line width to the terminal display width.

        context_line_count: t.Final = 2
        max_annotated_line_width: t.Final = 120
        truncation_marker: t.Final = '...'

        target_line_num = origin.line_num

        if RedactAnnotatedSourceContext.current(optional=True):
            return cls.error('content redacted', origin)

        if not target_line_num or target_line_num < 1:
            return cls.error(None, origin)  # message omitted since lack of line number is obvious from pos

        start_line_idx = max(0, (target_line_num - 1) - context_line_count)  # if near start of file
        target_col_num = origin.col_num

        try:
            with pathlib.Path(origin.path).open() as src:
                first_line = src.readline()
                lines = list(itertools.islice(itertools.chain((first_line,), src), start_line_idx, target_line_num))
        except Exception as ex:
            return cls.error(type(ex).__name__, origin)

        if is_encrypted(first_line):
            return cls.error('content encrypted', origin)

        if len(lines) != target_line_num - start_line_idx:
            return cls.error('file truncated', origin)

        annotated_source_lines = []

        line_label_width = len(str(target_line_num))
        max_src_line_len = max_annotated_line_width - line_label_width - 1

        usable_line_len = max_src_line_len

        for line_num, line in enumerate(lines, start_line_idx + 1):
            line = line.rstrip('\n')  # universal newline default mode on `open` ensures we'll never see anything but \n
            line = line.replace('\t', ' ')  # mixed tab/space handling is intentionally disabled since we're both format and display config agnostic

            if len(line) > max_src_line_len:
                line = line[: max_src_line_len - len(truncation_marker)] + truncation_marker
                usable_line_len = max_src_line_len - len(truncation_marker)

            annotated_source_lines.append(f'{str(line_num).rjust(line_label_width)}{" " if line else ""}{line}')

        if target_col_num and usable_line_len >= target_col_num >= 1:
            column_marker = f'column {target_col_num}'

            target_col_idx = target_col_num - 1

            if target_col_idx + 2 + len(column_marker) > max_src_line_len:
                column_marker = f'{" " * (target_col_idx - len(column_marker) - 1)}{column_marker} ^'
            else:
                column_marker = f'{" " * target_col_idx}^ {column_marker}'

            column_marker = f'{" " * line_label_width} {column_marker}'

            annotated_source_lines.append(column_marker)
        elif target_col_num is None:
            underline_length = len(annotated_source_lines[-1]) - line_label_width - 1
            annotated_source_lines.append(f'{" " * line_label_width} {"^" * underline_length}')

        return SourceContext(
            origin=origin,
            annotated_source_lines=annotated_source_lines,
            target_line=lines[-1].rstrip('\n'),  # universal newline default mode on `open` ensures we'll never see anything but \n
        )



from __future__ import annotations

import typing as t

from ansible.module_utils._internal import _ambient_context, _datatag

if t.TYPE_CHECKING:
    from ._engine import TemplateEngine, TemplateOptions


class TemplateContext(_ambient_context.AmbientContextBase):
    def __init__(
        self,
        *,
        template_value: t.Any,
        templar: TemplateEngine,
        options: TemplateOptions,
        stop_on_template: bool = False,
        _render_jinja_const_template: bool = False,
    ):
        self._template_value = template_value
        self._templar = templar
        self._options = options
        self._stop_on_template = stop_on_template
        self._parent_ctx = TemplateContext.current(optional=True)
        self._render_jinja_const_template = _render_jinja_const_template

    @property
    def is_top_level(self) -> bool:
        return not self._parent_ctx

    @property
    def template_value(self) -> t.Any:
        return self._template_value

    @property
    def templar(self) -> TemplateEngine:
        return self._templar

    @property
    def options(self) -> TemplateOptions:
        return self._options

    @property
    def stop_on_template(self) -> bool:
        return self._stop_on_template


class _OmitType:
    """
    A placeholder singleton used to dynamically omit items from a dict/list/tuple/set when the value is `Omit`.

    The Omit singleton value is accessible from all Ansible templating contexts via the Jinja global
    name `omit`. Item removal occurs during final recursive processing of template results. The singleton
    `Omit` placeholder value will be visible to plugins during templating. The only time a template result
    will include `Omit` outside a templating context is when the template renders to the scalar value `Omit`.
    """

    __slots__ = ()

    # DTFIX-MERGE: this keeps pickle happy, but not JSON/YAML for callbacks; just teach them about it?
    def __new__(cls):
        return Omit

    def __repr__(self):
        return "<<Omit>>"


Omit = object.__new__(_OmitType)

# DTFIX-MERGE: decide if these should be taggable; do we need to support other kinds of Undefineds, etc
_datatag._untaggable_types.add(_OmitType)

IGNORE_SCALAR_VAR_TYPES = {value for value in _datatag._ANSIBLE_ALLOWED_SCALAR_VAR_TYPES if not issubclass(value, str)}

PASS_THROUGH_SCALAR_VAR_TYPES = _datatag._ANSIBLE_ALLOWED_SCALAR_VAR_TYPES | {
    _OmitType,  # allow pass through of omit for later handling after top-level finalize completes
}



"""Public API for data tagging."""
from __future__ import annotations as _annotations

import datetime as _datetime
import typing as _t

from ._internal import _plugin_exec_context, _datatag
from ._internal._datatag import _tags

_T = _t.TypeVar('_T')


def deprecate_value(
    value: _T,
    msg: str,
    *,
    help_text: str | None = None,
    removal_date: str | _datetime.date | None = None,
    removal_version: str | None = None,
) -> _T:
    """
    Return `value` tagged with the given deprecation details.
    The types `None` and `bool` cannot be deprecated and are returned unmodified.
    Raises a `TypeError` if `value` is not a supported type.
    If `removal_date` is a string, it must be in the form `YYYY-MM-DD`.
    This function is only supported in contexts where an Ansible plugin/module is executing.
    """
    if isinstance(removal_date, str):
        # The `fromisoformat` method accepts other ISO 8601 formats than `YYYY-MM-DD` starting with Python 3.11.
        # That should be considered undocumented behavior of `deprecate_value` rather than an intentional feature.
        removal_date = _datetime.date.fromisoformat(removal_date)

    deprecated = _tags.Deprecated(
        msg=msg,
        help_text=help_text,
        removal_date=removal_date,
        removal_version=removal_version,
        plugin=_plugin_exec_context.PluginExecContext.get_current_plugin_info(),
    )

    return deprecated.tag(value)


def native_type_name(value: object | type, /) -> str:
    """Return the type name of `value`, substituting the native Python type name for internal tagged types."""
    return _datatag.AnsibleTagHelper.base_type(value).__name__



from __future__ import annotations as _annotations

import enum as _enum
import types as _types

from .common import json as _json
from ._internal import _serialization


class Direction(_enum.Enum):
    MODULE_TO_CONTROLLER = _enum.auto()
    CONTROLLER_TO_MODULE = _enum.auto()


def get_encoder(profile: str | _types.ModuleType, /) -> type[_json.AnsibleProfileJSONEncoder]:
    return _serialization.get_encoder_decoder(profile, _json.AnsibleProfileJSONEncoder)


def get_decoder(profile: str | _types.ModuleType, /) -> type[_json.AnsibleProfileJSONDecoder]:
    return _serialization.get_encoder_decoder(profile, _json.AnsibleProfileJSONDecoder)


def get_module_encoder(name: str, direction: Direction, /) -> type[_json.AnsibleProfileJSONEncoder]:
    return get_encoder(_serialization.get_module_serialization_profile_name(name, direction == Direction.CONTROLLER_TO_MODULE))


def get_module_decoder(name: str, direction: Direction, /) -> type[_json.AnsibleProfileJSONDecoder]:
    return get_decoder(_serialization.get_module_serialization_profile_name(name, direction == Direction.CONTROLLER_TO_MODULE))



from __future__ import annotations as _annotations

# deprecated: description='deprecate everything in this module', core_version='2.23'


class AnsibleMapping(dict):
    """Backwards compatibility type."""


class AnsibleUnicode(str):
    """Backwards compatibility type."""


class AnsibleSequence(list):
    """Backwards compatibility type."""


class AnsibleVaultEncryptedUnicode:
    """Backwards compatibility type."""



# Copyright: (c) 2018, Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

from ansible.utils.display import Display

Display().deprecated(msg='The ansible.parsing.ajson module has moved to ansible.module_utils.common.json', version='2.18')

# Imported for backwards compat
from ansible.module_utils.common.json import AnsibleJSONEncoder, AnsibleJSONDecoder  # pylint: disable=unused-import



# Copyright (c) 2024 Ansible Project
# Simplified BSD License (see licenses/simplified_bsd.txt or https://opensource.org/licenses/BSD-2-Clause)

"""Internal error handling logic for targets. Not for use on the controller."""

from __future__ import annotations

from . import _traceback
from ..common.messages import Detail, ErrorSummary


def create_error_summary(exception: BaseException) -> ErrorSummary:
    """Return an `ErrorDetail` created from the given exception."""
    return ErrorSummary(
        details=_create_error_details(exception),
        formatted_traceback=_traceback.maybe_extract_traceback(exception, _traceback.TracebackEvent.ERROR),
    )


def _create_error_details(exception: BaseException) -> tuple[Detail, ...]:
    """Return an `ErrorMessage` tuple created from the given exception."""
    target_exception: BaseException | None = exception
    error_details: list[Detail] = []

    while target_exception:
        error_details.append(Detail(msg=str(target_exception).strip()))

        target_exception = target_exception.__cause__

    return tuple(error_details)



#!/usr/bin/env python
from __future__ import annotations

import cProfile
import sys
import traceback

from ansible.module_utils.common.text.converters import to_text

target = sys.argv.pop(1)
myclass = "%sCLI" % target.capitalize()
module_name = f'ansible.cli.{target}'

try:
    # define cli
    mycli = getattr(__import__(module_name, fromlist=[myclass]), myclass)
except ImportError as e:
    if module_name in e.msg:
        raise Exception("Ansible sub-program not implemented: %s" % target) from None
    else:
        raise

try:
    args = [to_text(a, errors='surrogate_or_strict') for a in sys.argv]
except UnicodeError:
    sys.stderr.write(u"The full traceback was:\n\n%s" % to_text(traceback.format_exc()))
    sys.exit(u'Command line args are parsable to utf-8')

# init cli
cli = mycli(args)

print(cli.__class__.version_info(gitinfo=True))

# parse args
cli.parse()

# run
cProfile.run('cli.run()')



# -*- coding: utf-8 -*-
#
# Copyright (c) 2017 Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

import json
import typing as t

import pytest

from ansible.module_utils._internal._ansiballz import _ModulePluginWrapper
from ansible.module_utils._internal._plugin_exec_context import PluginExecContext
from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.serialization import get_module_decoder, Direction
from ansible.module_utils.common import warnings
from ansible.module_utils.common.messages import Detail, DeprecationSummary, WarningSummary, PluginInfo

pytestmark = pytest.mark.usefixtures("module_env_mocker")


@pytest.mark.parametrize('stdin', [{}], indirect=['stdin'])
def test_warn(am, capfd):

    am.warn('warning1')

    with pytest.raises(SystemExit):
        am.exit_json(warnings=['warning2'])
    out, err = capfd.readouterr()
    actual = json.loads(out, cls=get_module_decoder('legacy', Direction.MODULE_TO_CONTROLLER))['warnings']
    expected = [WarningSummary._from_details(Detail(msg=msg)) for msg in ['warning1', 'warning2']]
    assert actual == expected


@pytest.mark.parametrize('kwargs,plugin_name,stdin', (
    (dict(msg='deprecation1'), None, {}),
    (dict(msg='deprecation3', version='2.4'), None, {}),
    (dict(msg='deprecation4', date='2020-03-10'), None, {}),
    (dict(msg='deprecation5'), 'ansible.builtin.ping', {}),
    (dict(msg='deprecation7', version='2.4'), 'ansible.builtin.ping', {}),
    (dict(msg='deprecation8', date='2020-03-10'), 'ansible.builtin.ping', {}),
), indirect=['stdin'])
def test_deprecate(am: AnsibleModule, capfd, kwargs: dict[str, t.Any], plugin_name: str | None) -> None:
    plugin_info = PluginInfo(requested_name=plugin_name, resolved_name=plugin_name, type='module') if plugin_name else None
    executing_plugin = _ModulePluginWrapper(plugin_info) if plugin_info else None
    collection_name = plugin_name.rpartition('.')[0] if plugin_name else None

    with PluginExecContext.when(bool(executing_plugin), executing_plugin=executing_plugin):
        am.deprecate(**kwargs)

        assert warnings.get_deprecation_messages() == (dict(collection_name=collection_name, **kwargs),)

        with pytest.raises(SystemExit):
            am.exit_json(deprecations=['deprecation9', ('deprecation10', '2.4')])

    out, err = capfd.readouterr()

    output = json.loads(out, cls=get_module_decoder('legacy', Direction.MODULE_TO_CONTROLLER))

    assert ('warnings' not in output or output['warnings'] == [])

    msg = kwargs.pop('msg')

    assert output['deprecations'] == [
        DeprecationSummary._from_details(Detail(msg=msg), **kwargs, plugin=plugin_info),
        DeprecationSummary._from_details(Detail(msg='deprecation9'), plugin=plugin_info),
        DeprecationSummary._from_details(Detail(msg='deprecation10'), version='2.4', plugin=plugin_info),
    ]


@pytest.mark.parametrize('stdin', [{}], indirect=['stdin'])
def test_deprecate_without_list(am, capfd):
    with pytest.raises(SystemExit):
        am.exit_json(deprecations='Simple deprecation warning')

    out, err = capfd.readouterr()
    output = json.loads(out, cls=get_module_decoder('legacy', Direction.MODULE_TO_CONTROLLER))
    assert ('warnings' not in output or output['warnings'] == [])
    assert output['deprecations'] == [
        DeprecationSummary._from_details(Detail(msg='Simple deprecation warning')),
    ]


@pytest.mark.parametrize('stdin', [{}], indirect=['stdin'])
def test_deprecate_without_list_version_date_not_set(am, capfd):
    with pytest.raises(AssertionError) as ctx:
        am.deprecate('Simple deprecation warning', date='', version='')  # pylint: disable=ansible-deprecated-no-version
    assert ctx.value.args[0] == "implementation error -- version and date must not both be set"



#!/usr/bin/env python

import argparse
import datetime
import json
import io
import os
import pathlib
import pyclbr
import shutil
import subprocess
import sys
import tarfile

from jinja2 import Environment
from packaging.version import Version
from urllib.request import Request, urlopen

here = pathlib.Path(__file__).parent


FALLIBLE_VERSION = r'''
def fallible_version(func):
    def inner(*args, **kwargs):
        ret = func(*args, **kwargs)
        post_lines = [
            "",
            "",
            "Fallible experimental build {fallible_version}",
            "This build contains the following experimental features:",
            {features}
        ]
        return ret + "\n".join(post_lines)
    return inner


@fallible_version
'''.strip()


def fallible_version(version_str):
    today = datetime.date.today()
    if not version_str:
        # return a dev-tagged default version for today
        return f'{today.year}.{today.month}.{today.day}.dev0'

    parsed_ver = Version(version_str)

    # ensure value is a sane date, raise if not
    parsed_date = datetime.date(parsed_ver.major, parsed_ver.minor, parsed_ver.micro)

    min_ver = today - datetime.timedelta(days=365)
    max_ver = today + datetime.timedelta(days=7)

    # sanity check build year as current or prior year
    if not (min_ver < parsed_date < max_ver):
        raise argparse.ArgumentTypeError('fallible build date out of range (1 year back, 7 days ahead)')

    return parsed_ver.public


parser = argparse.ArgumentParser()
parser.add_argument('ansible_ref')
parser.add_argument('--github-token', default=os.getenv('GITHUB_TOKEN'))
parser.add_argument('--version', type=fallible_version, default=fallible_version(None),
                    help='a PEP440-compliant version string of the form YYYY.M.D')
group = parser.add_mutually_exclusive_group()
group.add_argument('--pr', action='append', type=int)
group.add_argument('--all', action='store_true', default=False)
args = parser.parse_args()


def build_request(url):
    headers = {}
    if args.github_token:
        headers['Authorization'] = f'token {args.github_token}'
    return Request(
        url,
        headers=headers,
    )


def get_fallible_prs():
    print('Fetching all fallible tagged PRs')
    url = (
        'https://api.github.com/repos/ansible/ansible/issues?'
        'labels=fallible&direction=asc&per_page=100'
    )
    for issue in json.load(urlopen(build_request(url))):
        yield issue['number']


def get_prs(prs):
    data = []
    for pr in prs:
        print(f'Fetching details for #{pr}')
        url = f'https://api.github.com/repos/ansible/ansible/pulls/{pr}'
        data.append(json.load(urlopen(build_request(url))))
    return data


def extract_version_tarball(ref, dest):
    print(f'Fetching GitHub tarball for {ref}')
    if ref[0].isdigit():
        ref_type = 'tags'
    else:
        ref_type = 'heads'
    url = (
        'https://github.com/ansible/ansible'
        f'/archive/refs/{ref_type}/{ref}.tar.gz'
    )

    t = tarfile.open(
        mode='r:gz',
        fileobj=io.BytesIO(urlopen(url).read())
    )
    for member in t.getmembers():
        member.name = member.name.partition('/')[2]
        t.extract(member, path=dest, filter='data')


def modify_version(prs):
    print('Patching --version')
    functions = pyclbr.readmodule_ex(
        'ansible.cli.arguments.option_helpers',
        path=[str(here / 'ansible' / 'lib')],
    )
    version = functions['version']
    path = pathlib.Path(version.file)

    newlines = []
    for pr in prs:
        newlines.append(
            f'            "  - {pr["title"]}: {pr["html_url"]}",'
        )

    lines = path.read_text().splitlines()
    start_lineno = version.lineno - 1
    lines[start_lineno:start_lineno] = [
        FALLIBLE_VERSION.format(fallible_version=args.version, features='\n'.join(newlines)),
    ]
    path.write_text('\n'.join(lines))


def get_patches(prs):
    for pr in prs:
        number = pr['number']
        print(f'Fetching patch for #{number}')
        url = f'https://github.com/ansible/ansible/pull/{number}.diff'
        yield number, urlopen(build_request(url)).read()


def apply_patch(pr, patch):
    print(f'Applying patch for #{pr}')
    p = subprocess.Popen(
        [
            'patch',
            '-p1'
        ],
        cwd=here / 'ansible',
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        stdin=subprocess.PIPE,
    )
    stdout, stderr = p.communicate(input=patch)
    if p.returncode:
        input(
            f'\nFailed to apply patch for {pr}. Resolve the merge conflicts '
            f'and then press ENTER:\n'
            f'\nstdout:\n{stdout.decode()}\nstderr:\n{stderr.decode()}\n'
        )
        print('Continuing')


def generate_fallible_version(fallible_version):
    print(f'Setting fallible version to {fallible_version}')
    version_path = (here / '_fallible_version.py')
    version_path.write_text(f'VERSION = "{fallible_version}"')


def generate_ansible_core_dist_info():
    print('Generating dist_info')
    p = subprocess.Popen(
        [
            sys.executable,
            'setup.py',
            'dist_info'
        ],
        cwd=here / 'ansible',
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        stdin=subprocess.PIPE,
    )
    p.communicate()


def generate_readme(prs):
    print('Generating FEATURES.md')
    features = (here.parent.parent.parent / 'FEATURES.md.j2').read_text()
    features_out = here.parent.parent.parent / 'FEATURES.md'
    e = Environment()
    t = e.from_string(features)
    features_out.write_text(t.render(prs=prs))


# cleanup
print('Cleaning up _vendor directory')
for path in here.glob('*'):
    if path.name in ('__init__.py', '_build.py'):
        continue
    if path.is_dir():
        shutil.rmtree(path)
    else:
        path.unlink()

ansible_dir = here / 'ansible'
ansible_dir.mkdir()

if args.all:
    prs = get_prs(get_fallible_prs())
else:
    prs = get_prs(args.pr or [])
patches = list(get_patches(prs))

extract_version_tarball(args.ansible_ref, ansible_dir)

for patch in patches:
    apply_patch(*patch)

modify_version(prs)

generate_ansible_core_dist_info()
generate_fallible_version(args.version)
generate_readme(prs)

# cleanup
print('Cleaning up patch files')
for path in here.glob('**/*.orig'):
    path.unlink()
for path in here.glob('**/*.rej'):
    path.unlink()



# Copyright: (c) 2021, Matt Martz <matt@sivel.net>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from ansible.__main__ import main

if __name__ == '__main__':
    main()



from __future__ import annotations

# DTFIX-MERGE: this module should probably move to module_utils/_internal

import sys
import dataclasses

# deprecated: description='typing.Self exists in Python 3.11+' python_version='3.10'
from ..compat import typing as t

from ansible.module_utils._internal._datatag import AnsibleSerializableDataclass

if sys.version_info >= (3, 10):
    # Using slots for reduced memory usage and improved performance.
    _dataclass_kwargs = dict(frozen=True, kw_only=True, slots=True)
else:
    # deprecated: description='always use dataclass slots and keyword-only args' python_version='3.9'
    _dataclass_kwargs = dict(frozen=True)


@dataclasses.dataclass(**_dataclass_kwargs)
class PluginInfo(AnsibleSerializableDataclass):
    """Information about a loaded plugin."""

    requested_name: str
    """The plugin name as requested, before resolving, which may be partially or fully qualified."""
    resolved_name: str
    """The resolved canonical plugin name; always fully-qualified for collection plugins."""
    type: str
    """The plugin type."""

    @property
    def description(self) -> str:
        """A brief description of the plugin, including name(s) and type."""
        name = repr(self.resolved_name)
        clarification = f' (requested as {self.requested_name!r})' if self.requested_name != self.resolved_name else ''

        if self.type in ("module", "modules"):
            # DTFIX-RELEASE: pluginloader or AnsiblePlugin needs a "type desc" property that doesn't suffer from legacy "inconsistencies" like this
            plugin_type = "module"
        elif self.type == "collection":
            # not a real plugin type, but used for tombstone errors generated by plugin loader
            plugin_type = self.type
        else:
            plugin_type = f'{self.type} plugin'

        return f'{name} {plugin_type}{clarification}'


@dataclasses.dataclass(**_dataclass_kwargs)
class Detail(AnsibleSerializableDataclass):
    """Message detail with optional source context and help text."""

    msg: str
    formatted_source_context: t.Optional[str] = None
    help_text: t.Optional[str] = None


@dataclasses.dataclass(**_dataclass_kwargs)
class SummaryBase(AnsibleSerializableDataclass):
    """Base class for an error/warning/deprecation summary with details (possibly derived from an exception __cause__ chain) and an optional traceback."""

    details: t.Tuple[Detail, ...]
    formatted_traceback: t.Optional[str] = None

    def format(self) -> str:
        """Returns a string representation of the warning details."""
        # DTFIX-RELEASE: should this borrow some of the message squashing features we use in get_chained_message?
        return ': '.join(detail.msg for detail in self.details)

    def _post_validate(self) -> None:
        if not self.details:
            raise ValueError(f'{type(self).__name__}.details cannot be empty')

    @classmethod
    def _from_details(cls, *details: Detail, formatted_traceback: t.Optional[str] = None, **kwargs) -> t.Self:
        """Utility factory method to avoid inline tuples."""
        return cls(details=details, formatted_traceback=formatted_traceback, **kwargs)


@dataclasses.dataclass(**_dataclass_kwargs)
class ErrorSummary(SummaryBase):
    """Error summary with details (possibly derived from an exception __cause__ chain) and an optional traceback."""


@dataclasses.dataclass(**_dataclass_kwargs)
class WarningSummary(SummaryBase):
    """Warning summary with details (possibly derived from an exception __cause__ chain) and an optional traceback."""


@dataclasses.dataclass(**_dataclass_kwargs)
class DeprecationSummary(WarningSummary):
    """Deprecation summary with details (possibly derived from an exception __cause__ chain) and an optional traceback."""

    version: t.Optional[str] = None
    date: t.Optional[str] = None
    plugin: t.Optional[PluginInfo] = None

    @property
    def collection_name(self) -> t.Optional[str]:
        if not self.plugin:
            return None

        parts = self.plugin.resolved_name.split('.')

        if len(parts) < 2:
            return None

        collection_name = '.'.join(parts[:2])

        # deprecated: description='enable the deprecation message for collection_name' core_version='2.23'
        # from ansible.module_utils.datatag import deprecate_value
        # collection_name = deprecate_value(collection_name, 'The `collection_name` property is deprecated.', removal_version='2.25')

        return collection_name

    def _as_simple_dict(self) -> t.Dict[str, t.Any]:
        """Returns a dictionary representation of the deprecation object in the format exposed to playbooks."""
        result = self._as_dict()
        result.pop('details')
        result.update(
            msg=self.format(),
            collection_name=self.collection_name,
        )

        return result



from __future__ import annotations

import typing as t

is_controller: bool = False
"""Set to True automatically when this module is imported into an Ansible controller context."""


def get_controller_serialize_map() -> dict[type, t.Callable]:
    """
    Called to augment serialization maps.
    This implementation is replaced with the one from ansible._internal in controller contexts.
    """
    return {}


def import_controller_module(_module_name: str, /) -> t.Any:
    """
    Called to conditionally import the named module in a controller context, otherwise returns `None`.
    This implementation is replaced with the one from ansible._internal in controller contexts.
    """
    return None



from __future__ import annotations

from ansible.module_utils.common.messages import PluginInfo
from ansible.plugins.lookup import LookupBase


class LookupModule(LookupBase):
    def run(self, terms, variables=None, **kwargs):
        return [PluginInfo(
            requested_name='requested_name',
            resolved_name='resolved_name',
            type='type',
        )]



import pathlib

from setuptools import setup

here = pathlib.Path(__file__).parent.resolve()

install_requires = (
    here / 'src' / 'fallible' / '_vendor' / 'ansible' / 'requirements.txt'
).read_text(encoding='utf-8').splitlines()

setup(
    install_requires=install_requires,
    entry_points={
        'console_scripts': [
            'fallible=fallible._vendor.ansible.lib.ansible.cli.adhoc:main',
            'fallible-config=fallible._vendor.ansible.lib.ansible.cli.config:main',
            'fallible-console=fallible._vendor.ansible.lib.ansible.cli.console:main',
            'fallible-doc=fallible._vendor.ansible.lib.ansible.cli.doc:main',
            'fallible-galaxy=fallible._vendor.ansible.lib.ansible.cli.galaxy:main',
            'fallible-inventory=fallible._vendor.ansible.lib.ansible.cli.inventory:main',
            'fallible-playbook=fallible._vendor.ansible.lib.ansible.cli.playbook:main',
            'fallible-pull=fallible._vendor.ansible.lib.ansible.cli.pull:main',
            'fallible-vault=fallible._vendor.ansible.lib.ansible.cli.vault:main',
            'fallible-connection=fallible._vendor.ansible.lib.ansible.cli.scripts.ansible_connection_cli_stub:main',
            'fallible-test=fallible._vendor.ansible.test.lib.ansible_test._util.target.cli.ansible_test_cli_stub:main',
        ]
    },
)



# -*- coding: utf-8 -*-
# Copyright (c) 2019 Ansible Project
# Simplified BSD License (see licenses/simplified_bsd.txt or https://opensource.org/licenses/BSD-2-Clause)

from __future__ import annotations

import datetime
import typing as t

from .._internal import _traceback, _plugin_exec_context
from ..common.messages import Detail, WarningSummary, DeprecationSummary

_UNSET = t.cast(t.Any, ...)


def warn(warning: str) -> None:
    """Record a warning to be returned with the module result."""
    _global_warnings[WarningSummary(
        details=(
            Detail(msg=warning),
        ),
        formatted_traceback=_traceback.maybe_capture_traceback(_traceback.TracebackEvent.WARNING),
    )] = None


def deprecate(msg: str, version: str | None = None, date: str | datetime.date | None = None, collection_name: str | None = _UNSET) -> None:
    """Record a deprecation warning to be returned with the module result."""
    if isinstance(date, datetime.date):
        date = str(date)

    # deprecated: description='enable the deprecation message for collection_name' core_version='2.23'
    # if collection_name is not _UNSET:
    #     deprecate('The `collection_name` argument to `deprecate` is deprecated.', version='2.25')

    _global_deprecations[DeprecationSummary(
        details=(
            Detail(msg=msg),
        ),
        formatted_traceback=_traceback.maybe_capture_traceback(_traceback.TracebackEvent.DEPRECATED),
        version=version,
        date=date,
        plugin=_plugin_exec_context.PluginExecContext.get_current_plugin_info(),
    )] = None


def get_warning_messages() -> tuple[str, ...]:
    """Return a tuple of warning messages accumulated over this run."""
    # DTFIX-RELEASE: add future deprecation comment
    return tuple(item.format() for item in _global_warnings)


def get_deprecation_messages() -> tuple[dict[str, t.Any], ...]:
    """Return a tuple of deprecation warning messages accumulated over this run."""
    # DTFIX-RELEASE: add future deprecation comment
    messages = [item._as_simple_dict() for item in _global_deprecations]

    for message in messages:
        message.pop('plugin', None)  # don't expose new data via legacy API

    return tuple(messages)


def get_warnings() -> list[WarningSummary]:
    """Return a list of warning messages accumulated over this run."""
    return list(_global_warnings)


def get_deprecations() -> list[DeprecationSummary]:
    """Return a list of deprecations accumulated over this run."""
    return list(_global_deprecations)


_global_warnings: dict[WarningSummary, object] = {}
"""Global, ordered, de-deplicated storage of acculumated warnings for the current module run."""

_global_deprecations: dict[DeprecationSummary, object] = {}
"""Global, ordered, de-deplicated storage of acculumated deprecations for the current module run."""



from __future__ import annotations

import dataclasses
import functools
import types
import typing as t

from collections import abc as c

from jinja2.environment import TemplateModule

from ansible.module_utils._internal._datatag import (
    AnsibleTaggedObject,
    _AnsibleTaggedDict,
    _AnsibleTaggedList,
    _NO_INSTANCE_STORAGE,
    _try_get_internal_tags_mapping,
    AnsibleTagHelper,
)

from ansible.utils.display import Display
from ansible.utils.sentinel import Sentinel
from ansible.errors import AnsibleVariableTypeError
from ansible._internal._errors._handler import Skippable
from ansible.vars.hostvars import HostVarsVars, HostVars

from ._access import AnsibleAccessContext
from ._jinja_common import Marker, _TemplateConfig
from ._utils import TemplateContext, PASS_THROUGH_SCALAR_VAR_TYPES

if t.TYPE_CHECKING:
    from ._engine import TemplateEngine, TemplateOptions

_ANSIBLE_LAZY_TEMPLATE_SLOTS = tuple(('_templar', '_template_options'))
_ITERATOR_TYPES: t.Final = (c.Iterator, c.ItemsView, c.KeysView, c.ValuesView, range)

display = Display()


class _AnsibleLazyTemplateMixin:
    __slots__ = _NO_INSTANCE_STORAGE

    # due to the way Jinja handles globals, we may encounter things like functions/methods in hooked getitem/getattr that
    # always pass through this mixin; we want to silently ignore those types
    # NB: additional values are added at runtime by other Python modules to avoid circular imports
    _ignore_types: t.ClassVar[set[type]] = (
        set(PASS_THROUGH_SCALAR_VAR_TYPES)
        | set(Marker.concrete_subclasses)  # vault marker is added later once it's defined
        | {
            HostVars,
            HostVarsVars,
            types.FunctionType,  # DTFIX-MERGE: global functions returned from Jinja globals __getitem__ def _lookup
            types.MethodType,  # DTFIX-MERGE: ?
            functools.partial,  # triggered by TaskExecutor lookup partial injection as a template local
            range,  # range is acceptable for historical reasons, but only within templating, it is always an error during finalization
            type,  # DTFIX-MERGE: this is a broad ignore for looking up `range` via `resolve_or_missing`; is there a better way?
            type(''.startswith),  # DTFIX-MERGE: builtin_function_or_method - is there a better way to include callables so we're not playing whack-a-mole?
            tuple,  # common native Python type, don't complain about it
            TemplateModule,  # the result of a Jinja `import` directive, just pass it through
        }
    )
    _ignore_types_tuple: t.ClassVar[tuple[type, ...]] = tuple(_ignore_types)

    _dispatch_types: dict[type, type[_AnsibleLazyTemplateMixin]] = {}  # populated by __init_subclass__
    _container_types: set[type] = set()  # populated by __init_subclass__

    _templar: TemplateEngine
    _template_options: TemplateOptions | None

    @classmethod
    def _register_ignore_types(cls, *args: type) -> None:
        """Register additional types to ignore."""
        cls._ignore_types.update(args)
        cls._ignore_types_tuple = tuple(cls._ignore_types)

    def __init_subclass__(cls, **kwargs) -> None:
        tagged_type = cls.__mro__[1]
        native_type = tagged_type.__mro__[1]

        cls._dispatch_types[native_type] = cls
        cls._dispatch_types[tagged_type] = cls
        cls._ignore_types.add(cls)
        cls._container_types.add(native_type)
        cls._empty_tags_as_native = False  # never revert to the native type when no tags remain

    def __init__(self, contents: t.Iterable | types.EllipsisType | _LazyIterator) -> None:
        ctx = TemplateContext.current()

        # DTFIX-MERGE: this doesn't feel quite right; what if we had a JinjaOperationContext or something that made the object we were getitem-ing on
        #  accessible to try_create so we could see if there's an active unmask_type_names to use/propagate?
        #  if we do use TemplateOptions, we need to verify the recursion/propgation behavior is actually what we want
        if ctx.options.unmask_type_names:
            from ._engine import TemplateOptions

            self._template_options = TemplateOptions(unmask_type_names=ctx.options.unmask_type_names)
        else:
            self._template_options = None

        if isinstance(contents, _LazyIterator):
            self._templar = contents.templar
        elif isinstance(contents, _AnsibleLazyTemplateMixin):
            self._templar = contents._templar
        else:
            self._templar = ctx.templar  # pylint: disable=assigning-non-slot  # slot defined in derived type

    def __reduce_ex__(self, protocol):
        raise NotImplementedError("Pickling of Ansible lazy objects is not permitted.")

    @staticmethod
    def _try_create(item: t.Any, auto_template: bool = True) -> t.Any:
        """
        If `item` is a container type which supports lazy access and/or templating, return a lazy wrapped version -- otherwise return it as-is.
        When returning as-is, a warning or error may be generated for unknown types.
        The `auto_template` argument should be set to `False` when `item` is sourced from a plugin instead of Ansible variable storage.
        This provides backwards compatibility and reduces lazy overhead, as plugins do not normally introduce templates.
        If a plugin needs to introduce templates, the plugin is responsible for invoking the templar and returning the result.
        """
        item_type = type(item)

        # Try to use exact type match first to determine which wrapper (if any) to apply; isinstance checks
        # are extremely expensive, so try to avoid them for our commonly-supported types.
        if (dispatcher := _AnsibleLazyTemplateMixin._dispatch_types.get(item_type)) is not None:
            # Create a generator that yields the elements of `item` wrapped in a `_LazyValue` wrapper.
            # The wrapper is used to signal to the lazy container that the value must be processed before being returned.
            # Values added to the lazy container later through other means will be returned as-is, without any special processing.
            lazy_values = dispatcher._lazy_values(item)

            if not auto_template:
                lazy_values = t.cast(c.Iterable, _LazyIterator(lazy_values, None))

            tags_mapping = _try_get_internal_tags_mapping(item)
            value = t.cast(AnsibleTaggedObject, dispatcher)._instance_factory(lazy_values, tags_mapping)

            return value

        with Skippable, _TemplateConfig.unsupported_variable_type_handler.handle(AnsibleVariableTypeError, skip_on_ignore=True):
            if item_type not in _AnsibleLazyTemplateMixin._ignore_types and not isinstance(item, _AnsibleLazyTemplateMixin._ignore_types_tuple):
                # DTFIX-MERGE: need a way to reliably ascend the template stack to describe the context for constant tuples
                #              we have a similar need in other locations for more than just constant tuples
                raise AnsibleVariableTypeError.from_value(obj=item)

        return item

    def _non_lazy_copy(self) -> t.Collection:
        """
        Return a non-lazy copy of this collection.
        Any remaining lazy wrapped values will be unwrapped without further processing.
        Tags on this instance will be preserved on the returned copy.
        """
        raise NotImplementedError()  # pragma: nocover

    @staticmethod
    def _lazy_values(values: t.Any, /) -> t.Iterable:
        """
        Return an iterable that wraps each of the given elements in a lazy wrapper.
        Only elements wrapped this way will receive lazy processing when retrieved from the collection.
        """
        raise NotImplementedError()  # pragma: nocover

    def _ensure_lazy_values(self) -> None:
        """Ensure that all values in the collection are properly lazified (required, abstract)."""
        raise NotImplementedError()  # pragma: nocover

    def _proxy_or_render_lazy_value(self, key: t.Any, value: t.Any) -> t.Any:
        """
        Ensure that the value is lazy-proxied or rendered, and if a key is provided, replace the original value with the result.
        """
        if type(value) is not _LazyValue:  # pylint: disable=unidiomatic-typecheck
            AnsibleAccessContext.current().access(value)
            return value

        original_value = value.value
        AnsibleAccessContext.current().access(original_value)

        if self._templar:
            new_value = self._templar.template(original_value, options=self._template_options)
        else:
            new_value = _AnsibleLazyTemplateMixin._try_create(original_value, auto_template=False)

        if new_value is not original_value:
            AnsibleAccessContext.current().access(new_value)

        if key is not NoKeySentinel:
            self._native_type.__setitem__(self, key, new_value)  # type: ignore  # pylint: disable=unnecessary-dunder-call

        return new_value


@t.final
@dataclasses.dataclass(frozen=True, slots=True)
class _LazyValue:
    """Wrapper around values to indicate lazy behavior has not yet been applied."""

    value: t.Any


class NoKeySentinel(Sentinel): ...


@t.final  # consumers of lazy collections rely heavily on the concrete types being final
class _AnsibleLazyTemplateDict(_AnsibleTaggedDict, _AnsibleLazyTemplateMixin):
    __slots__ = _ANSIBLE_LAZY_TEMPLATE_SLOTS

    def __init__(self, contents: t.Iterable | types.EllipsisType | _LazyIterator = ..., /, **kwargs):
        _AnsibleLazyTemplateMixin.__init__(self, contents)

        if isinstance(contents, types.EllipsisType):
            super().__init__(**kwargs)
        elif isinstance(contents, _AnsibleLazyTemplateDict):
            super().__init__(dict.items(contents), **kwargs)
        elif isinstance(contents, _LazyIterator):
            super().__init__(contents.iterator, **kwargs)
        else:
            super().__init__(contents, **kwargs)

    def get(self, key: t.Any, default: t.Any = None) -> t.Any:
        if (value := super().get(key, NoKeySentinel)) is NoKeySentinel:
            return default

        return self._proxy_or_render_lazy_value(key, value)

    def __getitem__(self, key: t.Any, /) -> t.Any:
        return self._proxy_or_render_lazy_value(key, super().__getitem__(key))

    def __str__(self):
        return str(self.copy()._native_copy())  # inefficient, but avoids mutating the current instance (to make debugging practical)

    def __repr__(self):
        return repr(self.copy()._native_copy())  # inefficient, but avoids mutating the current instance (to make debugging practical)

    def __iter__(self):
        # We're using the base implementation, but must override `__iter__` to skip `dict` fast-path copy, which would bypass lazy behavior.
        # See: https://github.com/python/cpython/blob/ffcc450a9b8b6927549b501eff7ac14abc238448/Objects/dictobject.c#L3861-L3864
        return super().__iter__()

    def _ensure_lazy_values(self):
        for key, value in super().items():
            if type(value) is not _LazyValue:  # pylint: disable=unidiomatic-typecheck
                super().__setitem__(key, _LazyValue(value))

    def setdefault(self, key, default=None, /) -> t.Any:
        if (value := self.get(key, NoKeySentinel)) is not NoKeySentinel:
            return value

        super().__setitem__(key, default)

        return default

    def items(self):
        for key, value in super().items():
            yield key, self._proxy_or_render_lazy_value(key, value)

    def values(self):
        for _key, value in self.items():
            yield value

    def pop(self, key, default=NoKeySentinel, /) -> t.Any:
        if (value := super().get(key, NoKeySentinel)) is NoKeySentinel:
            if default is NoKeySentinel:
                raise KeyError(key)

            return default

        value = self._proxy_or_render_lazy_value(NoKeySentinel, value)

        del self[key]

        return value

    def popitem(self) -> t.Any:
        try:
            key = next(reversed(self))
        except StopIteration:
            raise KeyError("popitem(): dictionary is empty")

        value = self._proxy_or_render_lazy_value(NoKeySentinel, self[key])

        del self[key]

        return key, value

    def _native_copy(self) -> dict:
        return dict(self.items())

    @staticmethod
    def _item_source(value: dict) -> dict | _LazyIterator:
        if isinstance(value, _AnsibleLazyTemplateDict):
            return _LazyIterator(dict.items(value), value._templar)

        return value

    def _yield_non_lazy_dict_items(self) -> t.Iterator[tuple[str, t.Any]]:
        """
        Delegate to the base collection items iterator to yield the raw contents.
        As of Python 3.13, generator functions are significantly faster than inline generator expressions.
        """
        for k, v in dict.items(self):
            yield k, v.value if type(v) is _LazyValue else v  # pylint: disable=unidiomatic-typecheck

    def _non_lazy_copy(self) -> dict:
        return AnsibleTagHelper.tag_copy(self, self._yield_non_lazy_dict_items(), value_type=dict)

    @staticmethod
    def _lazy_values(values: dict, /) -> t.Generator:
        yield from ((k, _LazyValue(v)) for k, v in values.items())

    @staticmethod
    def _proxy_or_render_other(other: t.Any | None) -> None:
        """Call `_proxy_or_render_lazy_values` if `other` is a lazy dict. Used internally by comparison methods."""
        if type(other) is _AnsibleLazyTemplateDict:  # pylint: disable=unidiomatic-typecheck
            other._proxy_or_render_lazy_values()

    def _proxy_or_render_lazy_values(self) -> None:
        """Ensure all `_LazyValue` wrapped values have been processed."""
        for _unused in self.values():
            pass

    def __eq__(self, other):
        self._proxy_or_render_lazy_values()
        self._proxy_or_render_other(other)
        return super().__eq__(other)

    def __ne__(self, other):
        self._proxy_or_render_lazy_values()
        self._proxy_or_render_other(other)
        return super().__ne__(other)

    def __or__(self, other):
        # Both sides end up going through _proxy_or_render_lazy_value, so there's no Templar preservation needed.
        # In the future this could be made more lazy when both Templar instances are the same, or if per-value Templar tracking was used.
        return _AnsibleLazyTemplateDict(super().__or__(other))

    def __ror__(self, other):
        # Both sides end up going through _proxy_or_render_lazy_value, so there's no Templar preservation needed.
        # In the future this could be made more lazy when both Templar instances are the same, or if per-value Templar tracking was used.
        return _AnsibleLazyTemplateDict(super().__ror__(other))

    def __deepcopy__(self, memo):
        # DTFIX-MERGE: implement lazy deep copy
        raise NotImplementedError("Deep copy of Ansible lazy types is not supported.")


class _LazyIterator:
    # DTFIX-MERGE: better way to smuggle this state around without this wrapper?
    def __init__(self, iterator: t.Iterable, templar: TemplateEngine | None) -> None:
        self.iterator = iterator
        self.templar = templar


@t.final  # consumers of lazy collections rely heavily on the concrete types being final
class _AnsibleLazyTemplateList(_AnsibleTaggedList, _AnsibleLazyTemplateMixin):
    __slots__ = _ANSIBLE_LAZY_TEMPLATE_SLOTS

    def __init__(self, contents: t.Iterable | types.EllipsisType | _LazyIterator = ..., /):
        _AnsibleLazyTemplateMixin.__init__(self, contents)

        if isinstance(contents, types.EllipsisType):
            super().__init__()
        elif isinstance(contents, _AnsibleLazyTemplateList):
            super().__init__(list.__iter__(contents))
        elif isinstance(contents, _LazyIterator):
            super().__init__(contents.iterator)
        else:
            super().__init__(contents)

    def __getitem__(self, key: t.SupportsIndex | slice, /) -> t.Any:
        if type(key) is slice:  # pylint: disable=unidiomatic-typecheck
            return _AnsibleLazyTemplateList(_LazyIterator(super().__getitem__(key), self._templar))

        return self._proxy_or_render_lazy_value(key, super().__getitem__(key))

    def __iter__(self):
        for key, value in enumerate(super().__iter__()):
            yield self._proxy_or_render_lazy_value(key, value)

    def pop(self, idx: t.SupportsIndex = -1, /) -> t.Any:
        if not self:
            raise IndexError('pop from empty list')

        try:
            value = self[idx]
        except IndexError:
            raise IndexError('pop index out of range')

        value = self._proxy_or_render_lazy_value(NoKeySentinel, value)

        del self[idx]

        return value

    def __str__(self):
        return str(self.copy()._native_copy())  # inefficient, but avoids mutating the current instance (to make debugging practical)

    def __repr__(self):
        return repr(self.copy()._native_copy())  # inefficient, but avoids mutating the current instance (to make debugging practical)

    @staticmethod
    def _item_source(value: list) -> list | _LazyIterator:
        if isinstance(value, _AnsibleLazyTemplateList):
            return _LazyIterator(list.__iter__(value), value._templar)

        return value

    def _yield_non_lazy_list_items(self):
        """
        Delegate to the base collection iterator to yield the raw contents.
        As of Python 3.13, generator functions are significantly faster than inline generator expressions.
        """
        for v in list.__iter__(self):
            yield v.value if type(v) is _LazyValue else v  # pylint: disable=unidiomatic-typecheck

    def _non_lazy_copy(self) -> list:
        return AnsibleTagHelper.tag_copy(self, self._yield_non_lazy_list_items(), value_type=list)

    @staticmethod
    def _lazy_values(values: list, /) -> t.Generator:
        yield from (_LazyValue(v) for v in values)

    def _ensure_lazy_values(self) -> None:
        for idx, value in enumerate(super().__iter__()):
            if type(value) is not _LazyValue:  # pylint: disable=unidiomatic-typecheck
                super().__setitem__(idx, _LazyValue(value))

    @staticmethod
    def _proxy_or_render_other(other: t.Any | None) -> None:
        """Call `_proxy_or_render_lazy_values` if `other` is a lazy list. Used internally by comparison methods."""
        if type(other) is _AnsibleLazyTemplateList:  # pylint: disable=unidiomatic-typecheck
            other._proxy_or_render_lazy_values()

    def _proxy_or_render_lazy_values(self) -> None:
        """Ensure all `_LazyValue` wrapped values have been processed."""
        for _unused in self:
            pass

    def __eq__(self, other):
        self._proxy_or_render_lazy_values()
        self._proxy_or_render_other(other)
        return super().__eq__(other)

    def __ne__(self, other):
        self._proxy_or_render_lazy_values()
        self._proxy_or_render_other(other)
        return super().__ne__(other)

    def __gt__(self, other):
        self._proxy_or_render_lazy_values()
        self._proxy_or_render_other(other)
        return super().__gt__(other)

    def __ge__(self, other):
        self._proxy_or_render_lazy_values()
        self._proxy_or_render_other(other)
        return super().__ge__(other)

    def __lt__(self, other):
        self._proxy_or_render_lazy_values()
        self._proxy_or_render_other(other)
        return super().__lt__(other)

    def __le__(self, other):
        self._proxy_or_render_lazy_values()
        self._proxy_or_render_other(other)
        return super().__le__(other)

    def __contains__(self, item):
        self._proxy_or_render_lazy_values()
        return super().__contains__(item)

    def __reversed__(self):
        for idx in range(self.__len__() - 1, -1, -1):
            yield self[idx]

    def __add__(self, other):
        if isinstance(other, _AnsibleLazyTemplateList) and self._templar is not other._templar:
            # with different templars, delazify both, new collection picks up the ambient templar
            self._proxy_or_render_lazy_values()

            return _AnsibleLazyTemplateList(super().__add__(other))  # other is delazified by its __iter__

        # For all other cases, the new list inherits our templar and all values stay lazy.
        # We use list.__add__ to avoid implementing all its error behavior.
        return _AnsibleLazyTemplateList(_LazyIterator(super().__add__(other), self._templar))

    def __radd__(self, other):
        if not (other_add := getattr(other, '__add__', None)):
            raise TypeError(f'unsupported operand type(s) for +: {type(other).__name__!r} and {type(self).__name__!r}') from None

        return _AnsibleLazyTemplateList(_LazyIterator(other_add(self), self._templar))

    def __mul__(self, other):
        return _AnsibleLazyTemplateList(_LazyIterator(super().__mul__(other), self._templar))

    def __rmul__(self, other):
        return _AnsibleLazyTemplateList(_LazyIterator(super().__rmul__(other), self._templar))

    def index(self, *args, **kwargs) -> int:
        self._proxy_or_render_lazy_values()
        return super().index(*args, **kwargs)

    def remove(self, *args, **kwargs) -> None:
        self._proxy_or_render_lazy_values()
        super().remove(*args, **kwargs)

    def sort(self, *args, **kwargs) -> None:
        self._proxy_or_render_lazy_values()
        super().sort(*args, **kwargs)

    def __deepcopy__(self, memo):
        # DTFIX-MERGE: implement lazy deep copy
        raise NotImplementedError("Deep copy of Ansible lazy types is not supported.")


def lazify_container(value: t.Any) -> t.Any:
    """
    If the given value is a supported container type, return its lazy version, otherwise return the value as-is.
    This is used to ensure that managed access and templating occur on args and kwargs to a callable, even if they were sourced from Jinja constants.

    Since both variable access and plugin output are already lazified, this mostly affects Jinja constant containers.
    However, plugins that directly invoke other plugins (e.g., `Environment.call_filter`) are another potential source of non-lazy containers.
    In these cases, templating will occur for trusted templates automatically upon access.

    Sets, tuples, and dictionary keys cannot be lazy, since their correct operation requires hashability and equality.
    These properties are mutually exclusive with the following lazy features:

    - managed access on encrypted strings - may raise errors on both operations when decryption fails
    - managed access on markers - must raise errors on both operations
    - templating - mutates values

    That leaves non-raising managed access as the only remaining feature, which is insufficient to warrant lazy support.
    """
    return _AnsibleLazyTemplateMixin._try_create(value)


def lazify_container_args(item: tuple) -> tuple:
    """Return the given args with values converted to lazy containers as needed."""
    return tuple(lazify_container(value) for value in item)


def lazify_container_kwargs(item: dict[str, t.Any]) -> dict[str, t.Any]:
    """Return the given kwargs with values converted to lazy containers as needed."""
    return {key: lazify_container(value) for key, value in item.items()}



from __future__ import annotations

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils._internal._datatag._tags import Deprecated


def main():
    mod = AnsibleModule(argument_spec={
        'sensitive_module_arg': dict(default='NO DISPLAY, sensitive arg to module', type='str', no_log=True),
    })

    result = {
        'good_key': 'good value',
        'deprecated_key': Deprecated(msg="`deprecated_key` is deprecated, don't use it!", removal_version='1.2.3').tag('deprecated value'),
        'sensitive_module_arg': mod.params['sensitive_module_arg'],
        'unmarked_template': '{{ ["me", "see", "not", "should"] | sort(reverse=true) | join(" ") }}',
        'changed': False
    }

    mod.exit_json(**result)


if __name__ == '__main__':
    main()



"""Runtime projections to provide template/var-visible views of objects that are not natively allowed in Ansible's type system."""

from __future__ import annotations

import dataclasses
import typing as t

from ansible.module_utils._internal import _traceback
from ansible.module_utils.common.messages import PluginInfo, ErrorSummary, WarningSummary, DeprecationSummary
from ansible.parsing.vault import EncryptedString, VaultHelper
from ansible.utils.display import Display

from ._jinja_common import VaultExceptionMarker
from .._errors import _captured, _utils

display = Display()


def plugin_info(value: PluginInfo) -> dict[str, str]:
    """Render PluginInfo as a dictionary."""
    return dataclasses.asdict(value)


def error_summary(value: ErrorSummary) -> str:
    """Render ErrorSummary as a formatted traceback for backward-compatibility with pre-2.19 TaskResult.exception."""
    return value.formatted_traceback or '(traceback unavailable)'


def warning_summary(value: WarningSummary) -> str:
    """Render WarningSummary as a simple message string for backward-compatibility with pre-2.19 TaskResult.warnings."""
    return value.format()


def deprecation_summary(value: DeprecationSummary) -> dict[str, t.Any]:
    """Render DeprecationSummary as dict values for backward-compatibility with pre-2.19 TaskResult.deprecations."""
    return value._as_simple_dict()


def encrypted_string(value: EncryptedString) -> str | VaultExceptionMarker:
    """Decrypt an encrypted string and return its value, or a VaultExceptionMarker if decryption fails."""
    try:
        return value._decrypt()
    except Exception as ex:
        return VaultExceptionMarker(
            ciphertext=VaultHelper.get_ciphertext(value, with_tags=True),
            reason=_utils.get_chained_message(ex),
            traceback=_traceback.maybe_extract_traceback(ex, _traceback.TracebackEvent.ERROR),
        )


_type_transform_mapping: dict[type, t.Callable[[t.Any], t.Any]] = {
    _captured.CapturedErrorSummary: error_summary,
    PluginInfo: plugin_info,
    ErrorSummary: error_summary,
    WarningSummary: warning_summary,
    DeprecationSummary: deprecation_summary,
    EncryptedString: encrypted_string,
}
"""This mapping is consulted by `Templar.template` to provide custom views of some objects."""



# Copyright: (c) 2017, Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

import atexit
import decimal
import configparser
import os
import os.path
import sys
import stat
import tempfile
import typing as t

from collections import namedtuple
from collections.abc import Mapping, Sequence
from jinja2.nativetypes import NativeEnvironment

from ansible.errors import AnsibleOptionsError, AnsibleError, AnsibleUndefinedConfigEntry, AnsibleRequiredOptionError
from ansible.module_utils.common.sentinel import Sentinel
from ansible.module_utils.common.text.converters import to_text, to_bytes, to_native
from ansible.module_utils.common.yaml import yaml_load
from ansible.module_utils.six import string_types
from ansible.module_utils.parsing.convert_bool import boolean
from ansible.parsing.quoting import unquote
from ansible.utils.path import cleanup_tmp_file, makedirs_safe, unfrackpath


Setting = namedtuple('Setting', 'name value origin type')

INTERNAL_DEFS = {'lookup': ('_terms',)}

GALAXY_SERVER_DEF = [
    ('url', True, 'str'),
    ('username', False, 'str'),
    ('password', False, 'str'),
    ('token', False, 'str'),
    ('auth_url', False, 'str'),
    ('api_version', False, 'int'),
    ('validate_certs', False, 'bool'),
    ('client_id', False, 'str'),
    ('client_secret', False, 'str'),
    ('timeout', False, 'int'),
]

# config definition fields
GALAXY_SERVER_ADDITIONAL = {
    'api_version': {'default': None, 'choices': [2, 3]},
    'validate_certs': {'cli': [{'name': 'validate_certs'}]},
    'timeout': {'cli': [{'name': 'timeout'}]},
    'token': {'default': None},
}


def _get_config_label(plugin_type: str, plugin_name: str, config: str) -> str:
    """Return a label for the given config."""
    entry = f'{config!r}'

    if plugin_type:
        entry += ' for'

        if plugin_name:
            entry += f' {plugin_name!r}'

        entry += f' {plugin_type} plugin'

    return entry


# FIXME: see if we can unify in module_utils with similar function used by argspec
def ensure_type(value, value_type, origin=None, origin_ftype=None):
    """ return a configuration variable with casting
    :arg value: The value to ensure correct typing of
    :kwarg value_type: The type of the value.  This can be any of the following strings:
        :boolean: sets the value to a True or False value
        :bool: Same as 'boolean'
        :integer: Sets the value to an integer or raises a ValueType error
        :int: Same as 'integer'
        :float: Sets the value to a float or raises a ValueType error
        :list: Treats the value as a comma separated list.  Split the value
            and return it as a python list.
        :none: Sets the value to None
        :path: Expands any environment variables and tilde's in the value.
        :tmppath: Create a unique temporary directory inside of the directory
            specified by value and return its path.
        :temppath: Same as 'tmppath'
        :tmp: Same as 'tmppath'
        :pathlist: Treat the value as a typical PATH string.  (On POSIX, this
            means comma separated strings.)  Split the value and then expand
            each part for environment variables and tildes.
        :pathspec: Treat the value as a PATH string. Expands any environment variables
            tildes's in the value.
        :str: Sets the value to string types.
        :string: Same as 'str'
    """

    errmsg = ''
    basedir = None
    if origin and os.path.isabs(origin) and os.path.exists(to_bytes(origin)):
        basedir = origin

    if value_type:
        value_type = value_type.lower()

    if value is not None:
        if value_type in ('boolean', 'bool'):
            value = boolean(value, strict=False)

        elif value_type in ('integer', 'int'):
            if not isinstance(value, int):
                try:
                    if (decimal_value := decimal.Decimal(value)) == (int_part := int(decimal_value)):
                        value = int_part
                    else:
                        errmsg = 'int'
                except decimal.DecimalException:
                    errmsg = 'int'

        elif value_type == 'float':
            if not isinstance(value, float):
                value = float(value)

        elif value_type == 'list':
            if isinstance(value, string_types):
                value = [unquote(x.strip()) for x in value.split(',')]
            elif not isinstance(value, Sequence):
                errmsg = 'list'

        elif value_type == 'none':
            if value == "None":
                value = None

            if value is not None:
                errmsg = 'None'

        elif value_type == 'path':
            if isinstance(value, string_types):
                value = resolve_path(value, basedir=basedir)
            else:
                errmsg = 'path'

        elif value_type in ('tmp', 'temppath', 'tmppath'):
            if isinstance(value, string_types):
                value = resolve_path(value, basedir=basedir)
                if not os.path.exists(value):
                    makedirs_safe(value, 0o700)
                prefix = 'ansible-local-%s' % os.getpid()
                value = tempfile.mkdtemp(prefix=prefix, dir=value)
                atexit.register(cleanup_tmp_file, value, warn=True)
            else:
                errmsg = 'temppath'

        elif value_type == 'pathspec':
            if isinstance(value, string_types):
                value = value.split(os.pathsep)

            if isinstance(value, Sequence):
                value = [resolve_path(x, basedir=basedir) for x in value]
            else:
                errmsg = 'pathspec'

        elif value_type == 'pathlist':
            if isinstance(value, string_types):
                value = [x.strip() for x in value.split(',')]

            if isinstance(value, Sequence):
                value = [resolve_path(x, basedir=basedir) for x in value]
            else:
                errmsg = 'pathlist'

        elif value_type in ('dict', 'dictionary'):
            if not isinstance(value, Mapping):
                errmsg = 'dictionary'

        elif value_type in ('str', 'string'):
            if isinstance(value, (string_types, bool, int, float, complex)):
                value = to_text(value, errors='surrogate_or_strict')
                if origin_ftype and origin_ftype == 'ini':
                    value = unquote(value)
            else:
                errmsg = 'string'

        # defaults to string type
        elif isinstance(value, (string_types)):
            value = to_text(value, errors='surrogate_or_strict')
            if origin_ftype and origin_ftype == 'ini':
                value = unquote(value)

        if errmsg:
            raise ValueError(f'Invalid type provided for {errmsg!r}: {value!r}')

    return to_text(value, errors='surrogate_or_strict', nonstring='passthru')


# FIXME: see if this can live in utils/path
def resolve_path(path, basedir=None):
    """ resolve relative or 'variable' paths """
    if '{{CWD}}' in path:  # allow users to force CWD using 'magic' {{CWD}}
        path = path.replace('{{CWD}}', os.getcwd())

    return unfrackpath(path, follow=False, basedir=basedir)


# FIXME: generic file type?
def get_config_type(cfile):

    ftype = None
    if cfile is not None:
        ext = os.path.splitext(cfile)[-1]
        if ext in ('.ini', '.cfg'):
            ftype = 'ini'
        elif ext in ('.yaml', '.yml'):
            ftype = 'yaml'
        else:
            raise AnsibleOptionsError("Unsupported configuration file extension for %s: %s" % (cfile, to_native(ext)))

    return ftype


# FIXME: can move to module_utils for use for ini plugins also?
def get_ini_config_value(p, entry):
    """ returns the value of last ini entry found """
    value = None
    if p is not None:
        try:
            value = p.get(entry.get('section', 'defaults'), entry.get('key', ''), raw=True)
        except Exception:  # FIXME: actually report issues here
            pass
    return value


def find_ini_config_file(warnings=None):
    """ Load INI Config File order(first found is used): ENV, CWD, HOME, /etc/ansible """
    # FIXME: eventually deprecate ini configs

    if warnings is None:
        # Note: In this case, warnings does nothing
        warnings = set()

    potential_paths = []

    # A value that can never be a valid path so that we can tell if ANSIBLE_CONFIG was set later
    # We can't use None because we could set path to None.
    # Environment setting
    path_from_env = os.getenv("ANSIBLE_CONFIG", Sentinel)
    if path_from_env is not Sentinel:
        path_from_env = unfrackpath(path_from_env, follow=False)
        if os.path.isdir(to_bytes(path_from_env)):
            path_from_env = os.path.join(path_from_env, "ansible.cfg")
        potential_paths.append(path_from_env)

    # Current working directory
    warn_cmd_public = False
    try:
        cwd = os.getcwd()
        perms = os.stat(cwd)
        cwd_cfg = os.path.join(cwd, "ansible.cfg")
        if perms.st_mode & stat.S_IWOTH:
            # Working directory is world writable so we'll skip it.
            # Still have to look for a file here, though, so that we know if we have to warn
            if os.path.exists(cwd_cfg):
                warn_cmd_public = True
        else:
            potential_paths.append(to_text(cwd_cfg, errors='surrogate_or_strict'))
    except OSError:
        # If we can't access cwd, we'll simply skip it as a possible config source
        pass

    # Per user location
    potential_paths.append(unfrackpath("~/.ansible.cfg", follow=False))

    # System location
    potential_paths.append("/etc/ansible/ansible.cfg")

    for path in potential_paths:
        b_path = to_bytes(path)
        if os.path.exists(b_path) and os.access(b_path, os.R_OK):
            break
    else:
        path = None

    # Emit a warning if all the following are true:
    # * We did not use a config from ANSIBLE_CONFIG
    # * There's an ansible.cfg in the current working directory that we skipped
    if path_from_env != path and warn_cmd_public:
        warnings.add(u"Ansible is being run in a world writable directory (%s),"
                     u" ignoring it as an ansible.cfg source."
                     u" For more information see"
                     u" https://docs.ansible.com/ansible/devel/reference_appendices/config.html#cfg-in-world-writable-dir"
                     % to_text(cwd))

    return path


def _add_base_defs_deprecations(base_defs):
    """Add deprecation source 'ansible.builtin' to deprecations in base.yml"""
    def process(entry):
        if 'deprecated' in entry:
            entry['deprecated']['collection_name'] = 'ansible.builtin'

    for dummy, data in base_defs.items():
        process(data)
        for section in ('ini', 'env', 'vars'):
            if section in data:
                for entry in data[section]:
                    process(entry)


class ConfigManager(object):

    DEPRECATED = []  # type: list[tuple[str, dict[str, str]]]
    WARNINGS = set()  # type: set[str]

    def __init__(self, conf_file=None, defs_file=None):

        self._base_defs = {}
        self._plugins = {}
        self._parsers = {}

        self._config_file = conf_file

        self._base_defs = self._read_config_yaml_file(defs_file or ('%s/base.yml' % os.path.dirname(__file__)))
        _add_base_defs_deprecations(self._base_defs)

        if self._config_file is None:
            # set config using ini
            self._config_file = find_ini_config_file(self.WARNINGS)

        # consume configuration
        if self._config_file:
            # initialize parser and read config
            self._parse_config_file()

        # ensure we always have config def entry
        self._base_defs['CONFIG_FILE'] = {'default': None, 'type': 'path'}

    def load_galaxy_server_defs(self, server_list):

        def server_config_def(section, key, required, option_type):
            config_def = {
                'description': 'The %s of the %s Galaxy server' % (key, section),
                'ini': [
                    {
                        'section': 'galaxy_server.%s' % section,
                        'key': key,
                    }
                ],
                'env': [
                    {'name': 'ANSIBLE_GALAXY_SERVER_%s_%s' % (section.upper(), key.upper())},
                ],
                'required': required,
                'type': option_type,
            }
            if key in GALAXY_SERVER_ADDITIONAL:
                config_def.update(GALAXY_SERVER_ADDITIONAL[key])
                # ensure we always have a default timeout
                if key == 'timeout' and 'default' not in config_def:
                    config_def['default'] = self.get_config_value('GALAXY_SERVER_TIMEOUT')

            return config_def

        if server_list:
            for server_key in server_list:
                if not server_key:
                    # To filter out empty strings or non truthy values as an empty server list env var is equal to [''].
                    continue

                # Config definitions are looked up dynamically based on the C.GALAXY_SERVER_LIST entry. We look up the
                # section [galaxy_server.<server>] for the values url, username, password, and token.
                defs = dict((k, server_config_def(server_key, k, req, value_type)) for k, req, value_type in GALAXY_SERVER_DEF)
                self.initialize_plugin_configuration_definitions('galaxy_server', server_key, defs)

    def template_default(self, value, variables):
        if isinstance(value, string_types) and (value.startswith('{{') and value.endswith('}}')) and variables is not None:
            # template default values if possible
            # NOTE: cannot use is_template due to circular dep
            try:
                # FIXME: This really should be using an immutable sandboxed native environment, not just native environment
                t = NativeEnvironment().from_string(value)
                value = t.render(variables)
            except Exception:
                pass  # not templatable
        return value

    def _read_config_yaml_file(self, yml_file):
        # TODO: handle relative paths as relative to the directory containing the current playbook instead of CWD
        # Currently this is only used with absolute paths to the `ansible/config` directory
        yml_file = to_bytes(yml_file)
        if os.path.exists(yml_file):
            with open(yml_file, 'rb') as config_def:
                return yaml_load(config_def) or {}
        raise AnsibleError(
            "Missing base YAML definition file (bad install?): %s" % to_native(yml_file))

    def _parse_config_file(self, cfile=None):
        """ return flat configuration settings from file(s) """
        # TODO: take list of files with merge/nomerge

        if cfile is None:
            cfile = self._config_file

        ftype = get_config_type(cfile)
        if cfile is not None:
            if ftype == 'ini':
                self._parsers[cfile] = configparser.ConfigParser(inline_comment_prefixes=(';',))
                with open(to_bytes(cfile), 'rb') as f:
                    try:
                        cfg_text = to_text(f.read(), errors='surrogate_or_strict')
                    except UnicodeError as e:
                        raise AnsibleOptionsError("Error reading config file(%s) because the config file was not utf8 encoded: %s" % (cfile, to_native(e)))
                try:
                    self._parsers[cfile].read_string(cfg_text)
                except configparser.Error as e:
                    raise AnsibleOptionsError("Error reading config file (%s): %s" % (cfile, to_native(e)))
            # FIXME: this should eventually handle yaml config files
            # elif ftype == 'yaml':
            #     with open(cfile, 'rb') as config_stream:
            #         self._parsers[cfile] = yaml_load(config_stream)
            else:
                raise AnsibleOptionsError("Unsupported configuration file type: %s" % to_native(ftype))

    def _find_yaml_config_files(self):
        """ Load YAML Config Files in order, check merge flags, keep origin of settings"""
        pass

    def get_plugin_options(self, plugin_type, name, keys=None, variables=None, direct=None):

        options = {}
        defs = self.get_configuration_definitions(plugin_type=plugin_type, name=name)
        for option in defs:
            options[option] = self.get_config_value(option, plugin_type=plugin_type, plugin_name=name, keys=keys, variables=variables, direct=direct)

        return options

    def get_plugin_vars(self, plugin_type, name):

        pvars = []
        for pdef in self.get_configuration_definitions(plugin_type=plugin_type, name=name).values():
            if 'vars' in pdef and pdef['vars']:
                for var_entry in pdef['vars']:
                    pvars.append(var_entry['name'])
        return pvars

    def get_plugin_options_from_var(self, plugin_type, name, variable):

        options = []
        for option_name, pdef in self.get_configuration_definitions(plugin_type=plugin_type, name=name).items():
            if 'vars' in pdef and pdef['vars']:
                for var_entry in pdef['vars']:
                    if variable == var_entry['name']:
                        options.append(option_name)
        return options

    def get_configuration_definition(self, name, plugin_type=None, plugin_name=None):

        ret = {}
        if plugin_type is None:
            ret = self._base_defs.get(name, None)
        elif plugin_name is None:
            ret = self._plugins.get(plugin_type, {}).get(name, None)
        else:
            ret = self._plugins.get(plugin_type, {}).get(plugin_name, {}).get(name, None)

        return ret

    def has_configuration_definition(self, plugin_type, name):

        has = False
        if plugin_type in self._plugins:
            has = (name in self._plugins[plugin_type])

        return has

    def get_configuration_definitions(self, plugin_type=None, name=None, ignore_private=False):
        """ just list the possible settings, either base or for specific plugins or plugin """

        ret = {}
        if plugin_type is None:
            ret = self._base_defs
        elif name is None:
            ret = self._plugins.get(plugin_type, {})
        else:
            ret = self._plugins.get(plugin_type, {}).get(name, {})

        if ignore_private:
            for cdef in list(ret.keys()):
                if cdef.startswith('_'):
                    del ret[cdef]
        return ret

    def _loop_entries(self, container, entry_list):
        """ repeat code for value entry assignment """

        value = None
        origin = None
        for entry in entry_list:
            name = entry.get('name')
            try:
                temp_value = container.get(name, None)
            except UnicodeEncodeError:
                self.WARNINGS.add(u'value for config entry {0} contains invalid characters, ignoring...'.format(to_text(name)))
                continue
            if temp_value is not None:  # only set if entry is defined in container
                value = temp_value
                origin = name

                # deal with deprecation of setting source, if used
                if 'deprecated' in entry:
                    self.DEPRECATED.append((entry['name'], entry['deprecated']))

        return value, origin

    def get_config_value(self, config, cfile=None, plugin_type=None, plugin_name=None, keys=None, variables=None, direct=None):
        """ wrapper """

        try:
            value, _drop = self.get_config_value_and_origin(config, cfile=cfile, plugin_type=plugin_type, plugin_name=plugin_name,
                                                            keys=keys, variables=variables, direct=direct)
        except AnsibleError:
            raise
        except Exception as ex:
            raise AnsibleError(f"Unhandled exception when retrieving {config!r}.") from ex
        return value

    def get_config_default(self, config: str, plugin_type: str | None = None, plugin_name: str | None = None) -> t.Any:
        """Return the default value for the specified configuration."""
        return self.get_configuration_definitions(plugin_type, plugin_name)[config]['default']

    def get_config_value_and_origin(self, config, cfile=None, plugin_type=None, plugin_name=None, keys=None, variables=None, direct=None):
        """ Given a config key figure out the actual value and report on the origin of the settings """
        if cfile is None:
            # use default config
            cfile = self._config_file

        if config == 'CONFIG_FILE':
            return cfile, ''

        # Note: sources that are lists listed in low to high precedence (last one wins)
        value = None
        origin = None
        origin_ftype = None

        defs = self.get_configuration_definitions(plugin_type=plugin_type, name=plugin_name)
        if config in defs:

            aliases = defs[config].get('aliases', [])

            # direct setting via plugin arguments, can set to None so we bypass rest of processing/defaults
            if direct:
                if config in direct:
                    value = direct[config]
                    origin = 'Direct'
                else:
                    direct_aliases = [direct[alias] for alias in aliases if alias in direct]
                    if direct_aliases:
                        value = direct_aliases[0]
                        origin = 'Direct'

            if value is None and variables and defs[config].get('vars'):
                # Use 'variable overrides' if present, highest precedence, but only present when querying running play
                value, origin = self._loop_entries(variables, defs[config]['vars'])
                origin = 'var: %s' % origin

            # use playbook keywords if you have em
            if value is None and defs[config].get('keyword') and keys:
                value, origin = self._loop_entries(keys, defs[config]['keyword'])
                origin = 'keyword: %s' % origin

            # automap to keywords
            # TODO: deprecate these in favor of explicit keyword above
            if value is None and keys:
                if config in keys:
                    value = keys[config]
                    keyword = config

                elif aliases:
                    for alias in aliases:
                        if alias in keys:
                            value = keys[alias]
                            keyword = alias
                            break

                if value is not None:
                    origin = 'keyword: %s' % keyword

            if value is None and 'cli' in defs[config]:
                # avoid circular import .. until valid
                from ansible import context
                value, origin = self._loop_entries(context.CLIARGS, defs[config]['cli'])
                origin = 'cli: %s' % origin

            # env vars are next precedence
            if value is None and defs[config].get('env'):
                value, origin = self._loop_entries(os.environ, defs[config]['env'])
                origin = 'env: %s' % origin

            # try config file entries next, if we have one
            if self._parsers.get(cfile, None) is None:
                self._parse_config_file(cfile)

            # attempt to read from config file
            if value is None and cfile is not None:
                ftype = get_config_type(cfile)
                if ftype and defs[config].get(ftype):
                    try:
                        for entry in defs[config][ftype]:
                            # load from config
                            if ftype == 'ini':
                                temp_value = get_ini_config_value(self._parsers[cfile], entry)
                            elif ftype == 'yaml':
                                raise AnsibleError('YAML configuration type has not been implemented yet')
                            else:
                                raise AnsibleError('Invalid configuration file type: %s' % ftype)

                            if temp_value is not None:
                                # set value and origin
                                value = temp_value
                                origin = cfile
                                origin_ftype = ftype
                                if 'deprecated' in entry:
                                    if ftype == 'ini':
                                        self.DEPRECATED.append(('[%s]%s' % (entry['section'], entry['key']), entry['deprecated']))
                                    else:
                                        raise AnsibleError('Unimplemented file type: %s' % ftype)

                    except Exception as e:
                        sys.stderr.write("Error while loading config %s: %s" % (cfile, to_native(e)))

            # set default if we got here w/o a value
            if value is None:
                if defs[config].get('required', False):
                    if not plugin_type or config not in INTERNAL_DEFS.get(plugin_type, {}):
                        raise AnsibleRequiredOptionError(f"Required config {_get_config_label(plugin_type, plugin_name, config)} not provided.")
                else:
                    origin = 'default'
                    value = self.template_default(defs[config].get('default'), variables)

            try:
                # ensure correct type, can raise exceptions on mismatched types
                value = ensure_type(value, defs[config].get('type'), origin=origin, origin_ftype=origin_ftype)
            except ValueError as ex:
                if origin.startswith('env:') and value == '':
                    # this is empty env var for non string so we can set to default
                    origin = 'default'
                    value = ensure_type(defs[config].get('default'), defs[config].get('type'), origin=origin, origin_ftype=origin_ftype)
                else:
                    raise AnsibleOptionsError(f'Config {_get_config_label(plugin_type, plugin_name, config)} from {origin!r} has an invalid value.') from ex

            # deal with restricted values
            if value is not None and 'choices' in defs[config] and defs[config]['choices'] is not None:
                invalid_choices = True  # assume the worst!
                if defs[config].get('type') == 'list':
                    # for a list type, compare all values in type are allowed
                    invalid_choices = not all(choice in defs[config]['choices'] for choice in value)
                else:
                    # these should be only the simple data types (string, int, bool, float, etc) .. ignore dicts for now
                    invalid_choices = value not in defs[config]['choices']

                if invalid_choices:

                    if isinstance(defs[config]['choices'], Mapping):
                        valid = ', '.join([to_text(k) for k in defs[config]['choices'].keys()])
                    elif isinstance(defs[config]['choices'], string_types):
                        valid = defs[config]['choices']
                    elif isinstance(defs[config]['choices'], Sequence):
                        valid = ', '.join([to_text(c) for c in defs[config]['choices']])
                    else:
                        valid = defs[config]['choices']

                    raise AnsibleOptionsError(f'Invalid value {value!r} for config {_get_config_label(plugin_type, plugin_name, config)}.',
                                              help_text=f'Valid values are: {valid}')

            # deal with deprecation of the setting
            if 'deprecated' in defs[config] and origin != 'default':
                self.DEPRECATED.append((config, defs[config].get('deprecated')))
        else:
            raise AnsibleUndefinedConfigEntry(f'No config definition exists for {_get_config_label(plugin_type, plugin_name, config)}.')

        return value, origin

    def initialize_plugin_configuration_definitions(self, plugin_type, name, defs):

        if plugin_type not in self._plugins:
            self._plugins[plugin_type] = {}

        self._plugins[plugin_type][name] = defs

    @staticmethod
    def get_deprecated_msg_from_config(dep_docs, include_removal=False, collection_name=None):

        removal = ''
        if include_removal:
            if 'removed_at_date' in dep_docs:
                removal = f"Will be removed in a release after {dep_docs['removed_at_date']}\n\t"
            elif collection_name:
                removal = f"Will be removed in: {collection_name} {dep_docs['removed_in']}\n\t"
            else:
                removal = f"Will be removed in: Ansible {dep_docs['removed_in']}\n\t"

        # TODO: choose to deprecate either singular or plural
        alt = dep_docs.get('alternatives', dep_docs.get('alternative', 'none'))
        return f"Reason: {dep_docs['why']}\n\t{removal}Alternatives: {alt}"



from __future__ import annotations

import contextlib
import re
import typing as t

from ansible.utils.display import _DeferredWarningContext


@contextlib.contextmanager
def emits_deprecation_warning(match: str) -> t.Iterator[None]:
    """Assert that the code within the context manager body emits a deprecation warning whose formatted output matches the supplied regex."""
    with _DeferredWarningContext(variables=dict(ansible_deprecation_warnings=True)) as ctx:
        yield

    warnings = ctx.get_deprecation_warnings()

    assert re.search(match, str(warnings))
    assert len(warnings) == 1
